from dbt.tests.adapter.ephemeral.test_ephemeral import BaseEphemeralMulti
from dbt.tests.util import run_dbt, check_relations_equal


class TestEphemeralMultiSnowflake(BaseEphemeralMulti):
    def test_ephemeral_multi(self, project):
        run_dbt(["seed"])
        results = run_dbt(["run"])
        assert len(results) == 3
        check_relations_equal(
            project.adapter, ["SEED", "DEPENDENT", "DOUBLE_DEPENDENT", "SUPER_DEPENDENT"]
        )



import pytest
from dbt.tests.adapter.aliases.test_aliases import BaseAliases

MACROS__SNOWFLAKE_CAST_SQL = """
{% macro snowflake__string_literal(s) %}
    cast('{{ s }}' as string)
{% endmacro %}
"""

MACROS__EXPECT_VALUE_SQL = """
-- cross-db compatible test, similar to accepted_values

{% test expect_value(model, field, value) %}

select *
from {{ model }}
where {{ field }} != '{{ value }}'

{% endtest %}
"""


class TestAliasesSnowflake(BaseAliases):
    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "snowflake_cast.sql": MACROS__SNOWFLAKE_CAST_SQL,
            "expect_value.sql": MACROS__EXPECT_VALUE_SQL,
        }



import os
import sys
import tempfile
from typing import Generator

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
import pytest

from dbt.adapters.snowflake.auth import private_key_from_file, private_key_from_string


PASSPHRASE = "password1234"


def serialize(private_key: rsa.RSAPrivateKey) -> bytes:
    return private_key.private_bytes(
        serialization.Encoding.DER,
        serialization.PrivateFormat.PKCS8,
        serialization.NoEncryption(),
    )


@pytest.fixture(scope="session")
def private_key() -> rsa.RSAPrivateKey:
    return rsa.generate_private_key(public_exponent=65537, key_size=2048)


@pytest.fixture(scope="session")
def private_key_string(private_key) -> str:
    private_key_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.BestAvailableEncryption(PASSPHRASE.encode()),
    )
    return private_key_bytes.decode("utf-8")


@pytest.fixture(scope="session")
def private_key_file(private_key) -> Generator[str, None, None]:
    private_key_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.BestAvailableEncryption(PASSPHRASE.encode()),
    )
    file = tempfile.NamedTemporaryFile()
    file.write(private_key_bytes)
    file.seek(0)
    yield file.name
    file.close()


def test_private_key_from_string_pem(private_key_string, private_key):
    assert isinstance(private_key_string, str)
    calculated_private_key = private_key_from_string(private_key_string, PASSPHRASE)
    assert serialize(calculated_private_key) == serialize(private_key)


@pytest.mark.skipif(sys.platform == "win32", reason="permission issues on Windows")
def test_private_key_from_file(private_key_file, private_key):
    assert os.path.exists(private_key_file)
    calculated_private_key = private_key_from_file(private_key_file, PASSPHRASE)
    assert serialize(calculated_private_key) == serialize(private_key)



import os

from dbt.tests.util import run_dbt
import pytest


class TestKeyPairAuth:
    @pytest.fixture(scope="class", autouse=True)
    def dbt_profile_target(self):
        return {
            "type": "snowflake",
            "threads": 4,
            "account": os.getenv("SNOWFLAKE_TEST_ACCOUNT"),
            "user": os.getenv("SNOWFLAKE_TEST_USER"),
            "private_key": os.getenv("SNOWFLAKE_TEST_PRIVATE_KEY"),
            "private_key_passphrase": os.getenv("SNOWFLAKE_TEST_PRIVATE_KEY_PASSPHRASE"),
            "database": os.getenv("SNOWFLAKE_TEST_DATABASE"),
            "warehouse": os.getenv("SNOWFLAKE_TEST_WAREHOUSE"),
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {"my_model.sql": "select 1 as id"}

    def test_connection(self, project):
        run_dbt()



from dbt.tests.adapter.store_test_failures_tests import basic
from dbt.tests.adapter.store_test_failures_tests.test_store_test_failures import (
    TestStoreTestFailures,
)


class TestSnowflakeStoreTestFailures(TestStoreTestFailures):
    pass


class TestStoreTestFailuresAsInteractions(basic.StoreTestFailuresAsInteractions):
    pass


class TestStoreTestFailuresAsProjectLevelOff(basic.StoreTestFailuresAsProjectLevelOff):
    pass


class TestStoreTestFailuresAsProjectLevelView(basic.StoreTestFailuresAsProjectLevelView):
    pass


class TestStoreTestFailuresAsGeneric(basic.StoreTestFailuresAsGeneric):
    pass


class TestStoreTestFailuresAsProjectLevelEphemeral(basic.StoreTestFailuresAsProjectLevelEphemeral):
    pass


class TestStoreTestFailuresAsExceptions(basic.StoreTestFailuresAsExceptions):
    pass



import os
import pytest
import logging
from importlib import reload
from unittest.mock import Mock, patch, call
import multiprocessing
from dbt.adapters.exceptions.connection import FailedToConnectError
import dbt.adapters.snowflake.connections as connections
import dbt.adapters.events.logging


class TestSnowflakeLogging:
    """Test suite for snowflake logging configuration"""

    def test_connections_sets_debug_logs_when_env_var_present(self, monkeypatch):
        """Test that setting DBT_SNOWFLAKE_CONNECTOR_DEBUG_LOGGING enables DEBUG level logging"""
        log_mock = Mock()
        logger_mock = Mock(return_value=log_mock)

        monkeypatch.setattr(dbt.adapters.events.logging, "AdapterLogger", logger_mock)
        monkeypatch.setattr(os, "environ", {"DBT_SNOWFLAKE_CONNECTOR_DEBUG_LOGGING": "true"})

        reload(connections)

        # Verify AdapterLogger was created with correct name
        logger_mock.assert_called_once_with("Snowflake")

        # Verify all three dependency loggers are configured with DEBUG level
        expected_debug_calls = [
            call("Setting snowflake.connector to DEBUG (file logging only)"),
            call("Setting botocore to DEBUG (file logging only)"),
            call("Setting boto3 to DEBUG (file logging only)"),
        ]
        log_mock.debug.assert_has_calls(expected_debug_calls)

        expected_level_calls = [
            call("snowflake.connector", "DEBUG"),
            call("botocore", "DEBUG"),
            call("boto3", "DEBUG"),
        ]
        log_mock.set_adapter_dependency_log_level.assert_has_calls(expected_level_calls)

    def test_connections_sets_error_logs_when_env_var_absent(self, monkeypatch):
        """Test that absence of DBT_SNOWFLAKE_CONNECTOR_DEBUG_LOGGING sets ERROR level logging"""
        log_mock = Mock()
        logger_mock = Mock(return_value=log_mock)

        monkeypatch.setattr(dbt.adapters.events.logging, "AdapterLogger", logger_mock)
        # Explicitly set environment without the debug flag
        monkeypatch.setattr(os, "environ", {})

        with patch("logging.getLogger") as mock_get_logger:
            reload(connections)

        # Verify AdapterLogger was created
        logger_mock.assert_called_once_with("Snowflake")

        # Verify all three dependency loggers are configured with ERROR level
        expected_debug_calls = [
            call("Setting snowflake.connector to ERROR (file logging only)"),
            call("Setting botocore to ERROR (file logging only)"),
            call("Setting boto3 to ERROR (file logging only)"),
        ]

        log_mock.debug.assert_has_calls(expected_debug_calls)

        expected_level_calls = [
            call("snowflake.connector", "ERROR"),
            call("botocore", "ERROR"),
            call("boto3", "ERROR"),
        ]
        log_mock.set_adapter_dependency_log_level.assert_has_calls(expected_level_calls)

    def test_connections_handles_various_env_var_values(self, monkeypatch):
        """Test that any truthy value for the env var enables debug logging"""
        test_values = ["1", "True", "YES", "on", "random"]

        for value in test_values:
            log_mock = Mock()
            logger_mock = Mock(return_value=log_mock)

            monkeypatch.setattr(dbt.adapters.events.logging, "AdapterLogger", logger_mock)
            monkeypatch.setattr(os, "environ", {"DBT_SNOWFLAKE_CONNECTOR_DEBUG_LOGGING": value})

            with patch("logging.getLogger"):
                reload(connections)

            # Should set DEBUG level for all loggers (fallback for non-INFO/DEBUG values)
            expected_level_calls = [
                call("snowflake.connector", "DEBUG"),
                call("botocore", "DEBUG"),
                call("boto3", "DEBUG"),
            ]
            log_mock.set_adapter_dependency_log_level.assert_has_calls(expected_level_calls)

    def test_connections_handles_case_insensitive_values(self, monkeypatch):
        """Test that DEBUG and INFO values are case insensitive"""
        test_cases = [
            ("debug", "DEBUG"),
            ("DEBUG", "DEBUG"),
            ("info", "INFO"),
            ("INFO", "INFO"),
            ("Debug", "DEBUG"),
            ("Info", "INFO"),
        ]

        for env_value, expected_level in test_cases:
            log_mock = Mock()
            logger_mock = Mock(return_value=log_mock)

            monkeypatch.setattr(dbt.adapters.events.logging, "AdapterLogger", logger_mock)
            monkeypatch.setattr(
                os, "environ", {"DBT_SNOWFLAKE_CONNECTOR_DEBUG_LOGGING": env_value}
            )

            with patch("logging.getLogger"):
                reload(connections)

            # Should set the expected level for all loggers
            expected_level_calls = [
                call("snowflake.connector", expected_level),
                call("botocore", expected_level),
                call("boto3", expected_level),
            ]
            log_mock.set_adapter_dependency_log_level.assert_has_calls(expected_level_calls)

    def test_setup_snowflake_logging_with_info_level(self):
        """Test setup_snowflake_logging with INFO level"""
        with patch("dbt.adapters.snowflake.connections.logger") as mock_logger:
            with patch("logging.getLogger") as mock_get_logger:
                mock_package_logger = Mock()
                mock_get_logger.return_value = mock_package_logger

                # Test INFO level
                connections.setup_snowflake_logging("INFO")

                # Verify debug messages (always say "DEBUG (file logging only)")
                expected_debug_calls = [
                    call("Setting snowflake.connector to INFO (file logging only)"),
                    call("Setting botocore to INFO (file logging only)"),
                    call("Setting boto3 to INFO (file logging only)"),
                ]
                mock_logger.debug.assert_has_calls(expected_debug_calls)

                # Verify correct level is set (INFO in this case)
                expected_level_calls = [
                    call("snowflake.connector", "INFO"),
                    call("botocore", "INFO"),
                    call("boto3", "INFO"),
                ]
                mock_logger.set_adapter_dependency_log_level.assert_has_calls(expected_level_calls)

    def test_expected_logger_names_are_configured(self):
        """Test that exactly the expected logger names are configured"""
        expected_loggers = ["snowflake.connector", "botocore", "boto3"]

        with patch("dbt.adapters.snowflake.connections.logger") as mock_logger:
            with patch("logging.getLogger") as mock_get_logger:
                connections.setup_snowflake_logging("DEBUG")

                # Extract the logger names from the calls
                actual_logger_names = [
                    call[0][0]
                    for call in mock_logger.set_adapter_dependency_log_level.call_args_list
                ]

                assert actual_logger_names == expected_loggers
                assert len(actual_logger_names) == 3


# Keep existing tests for backward compatibility
def test_connections_sets_logs_in_response_to_env_var(monkeypatch):
    """Test that setting the DBT_SNOWFLAKE_CONNECTOR_DEBUG_LOGGING environment variable happens on import"""
    log_mock = Mock()
    monkeypatch.setattr(dbt.adapters.events.logging, "AdapterLogger", Mock(return_value=log_mock))
    monkeypatch.setattr(os, "environ", {"DBT_SNOWFLAKE_CONNECTOR_DEBUG_LOGGING": "true"})
    reload(connections)

    assert log_mock.debug.call_count == 3
    assert log_mock.set_adapter_dependency_log_level.call_count == 3


def test_connections_does_not_set_logs_in_response_to_env_var(monkeypatch):
    log_mock = Mock()
    monkeypatch.setattr(dbt.adapters.events.logging, "AdapterLogger", Mock(return_value=log_mock))
    reload(connections)

    assert log_mock.debug.call_count == 3  # Debug messages are always logged
    assert (
        log_mock.set_adapter_dependency_log_level.call_count == 3
    )  # But with ERROR level (default)


def test_connnections_credentials_replaces_underscores_with_hyphens():
    credentials = {
        "account": "account_id_with_underscores",
        "user": "user",
        "password": "password",
        "database": "database",
        "warehouse": "warehouse",
        "schema": "schema",
    }
    creds = connections.SnowflakeCredentials(**credentials)
    assert creds.account == "account-id-with-underscores"


def test_snowflake_oauth_expired_token_raises_error():
    credentials = {
        "account": "test_account",
        "user": "test_user",
        "authenticator": "oauth",
        "token": "expired_or_incorrect_token",
        "database": "database",
        "schema": "schema",
    }

    mp_context = multiprocessing.get_context("spawn")
    mock_credentials = connections.SnowflakeCredentials(**credentials)

    with patch.object(
        connections.SnowflakeConnectionManager,
        "open",
        side_effect=FailedToConnectError(
            "This error occurs when authentication has expired. "
            "Please reauth with your auth provider."
        ),
    ):

        adapter = connections.SnowflakeConnectionManager(mock_credentials, mp_context)

        with pytest.raises(FailedToConnectError):
            adapter.open()



from dbt.adapters.snowflake.relation import SnowflakeRelation
from dbt.adapters.snowflake.relation_configs import SnowflakeRelationType


def test_renameable_relation():
    relation = SnowflakeRelation.create(
        database="my_db",
        schema="my_schema",
        identifier="my_table",
        type=SnowflakeRelationType.Table,
    )
    assert relation.renameable_relations == frozenset(
        {
            SnowflakeRelationType.Table,
            SnowflakeRelationType.View,
            SnowflakeRelationType.DynamicTable,
        }
    )



import pytest

from dbt.tests.util import run_dbt, run_dbt_and_capture


SEED = """
id
0
1
2
""".strip()


TABLE = """
{{ config(materialized="table") }}
select * from {{ ref('my_seed') }}
"""


VIEW = """
{{ config(materialized="view") }}
select * from {{ ref('my_seed') }}
"""


MACRO__GET_CREATE_BACKUP_SQL = """
{% macro test__get_create_backup_sql(database, schema, identifier, relation_type) -%}
    {%- set relation = adapter.Relation.create(database=database, schema=schema, identifier=identifier, type=relation_type) -%}
    {% call statement('test__get_create_backup_sql') -%}
        {{ get_create_backup_sql(relation) }}
    {%- endcall %}
{% endmacro %}"""


MACRO__GET_RENAME_INTERMEDIATE_SQL = """
{% macro test__get_rename_intermediate_sql(database, schema, identifier, relation_type) -%}
    {%- set relation = adapter.Relation.create(database=database, schema=schema, identifier=identifier, type=relation_type) -%}
    {% call statement('test__get_rename_intermediate_sql') -%}
        {{ get_rename_intermediate_sql(relation) }}
    {%- endcall %}
{% endmacro %}"""


class RelationOperation:
    @pytest.fixture(scope="class")
    def seeds(self):
        yield {"my_seed.csv": SEED}

    @pytest.fixture(scope="class")
    def models(self):
        yield {
            "my_table.sql": TABLE,
            "my_table__dbt_tmp.sql": TABLE,
            "my_view.sql": VIEW,
            "my_view__dbt_tmp.sql": VIEW,
        }

    @pytest.fixture(scope="class")
    def macros(self):
        yield {
            "test__get_create_backup_sql.sql": MACRO__GET_CREATE_BACKUP_SQL,
            "test__get_rename_intermediate_sql.sql": MACRO__GET_RENAME_INTERMEDIATE_SQL,
        }

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])

    def assert_operation(self, project, operation, args, expected_statement):
        results, logs = run_dbt_and_capture(
            ["--debug", "run-operation", operation, "--args", str(args)]
        )
        assert len(results) == 1
        assert expected_statement in logs



import base64
import sys
from typing import Optional

if sys.version_info < (3, 9):
    from functools import lru_cache

    cache = lru_cache(maxsize=None)
else:
    from functools import cache

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey


@cache
def private_key_from_string(
    private_key_string: str, passphrase: Optional[str] = None
) -> RSAPrivateKey:

    if passphrase:
        encoded_passphrase = passphrase.encode()
    else:
        encoded_passphrase = None

    if private_key_string.startswith("-"):
        return serialization.load_pem_private_key(
            data=bytes(private_key_string, "utf-8"),
            password=encoded_passphrase,
            backend=default_backend(),
        )
    return serialization.load_der_private_key(
        data=base64.b64decode(private_key_string),
        password=encoded_passphrase,
        backend=default_backend(),
    )


@cache
def private_key_from_file(
    private_key_path: str, passphrase: Optional[str] = None
) -> RSAPrivateKey:

    if passphrase:
        encoded_passphrase = passphrase.encode()
    else:
        encoded_passphrase = None

    with open(private_key_path, "rb") as file:
        private_key_bytes = file.read()

    return serialization.load_pem_private_key(
        data=private_key_bytes,
        password=encoded_passphrase,
        backend=default_backend(),
    )



from dbt.tests.util import AnyString, AnyInteger


def snowflake_stats():
    return {
        "has_stats": {
            "id": "has_stats",
            "label": "Has Stats?",
            "value": True,
            "description": "Indicates whether there are statistics for this table",
            "include": False,
        },
        "bytes": {
            "id": "bytes",
            "label": "Approximate Size",
            "value": AnyInteger(),
            "description": "Approximate size of the table as reported by Snowflake",
            "include": True,
        },
        "last_modified": {
            "id": "last_modified",
            "label": "Last Modified",
            "value": AnyString(),
            "description": "The timestamp for last update/change",
            "include": True,
        },
        "row_count": {
            "id": "row_count",
            "label": "Row Count",
            "value": 1.0,
            "description": "An approximate count of rows in this table",
            "include": True,
        },
    }



from types import SimpleNamespace
from typing import Dict, Optional

import pytest

from dbt.adapters.contracts.relation import RelationConfig
from dbt.artifacts.resources.v1.saved_query import ExportConfig, ExportDestinationType
from dbt.adapters.snowflake.parse_model import (
    base_location,
    cluster_by,
    partition_by,
    catalog_name,
)


def make_model(config: Dict[str, Optional[str]]) -> RelationConfig:
    return SimpleNamespace(  # type:ignore
        database="fake_database",
        schema="fake_schema",
        identifier="fake_table",
        config=config,
    )


@pytest.mark.parametrize(
    "config,expected",
    [
        (
            {"fake_attr": "fake_value"},  # we check if not model.config
            "_dbt/fake_schema/fake_table",
        ),
        (
            {"base_location_root": None, "base_location_subpath": None},
            "_dbt/fake_schema/fake_table",
        ),
        (
            {"base_location_root": "root_path", "base_location_subpath": "subpath"},
            "root_path/fake_schema/fake_table/subpath",
        ),
        (
            {"base_location_subpath": "subpath"},
            "_dbt/fake_schema/fake_table/subpath",
        ),
        (
            {"base_location_root": "root_path"},
            "root_path/fake_schema/fake_table",
        ),
    ],
)
def test_base_location(config, expected):
    model = make_model(config)
    assert base_location(model) == expected


@pytest.mark.parametrize(
    "config,expected",
    [
        ({"cluster_by": "fake_value"}, "fake_value"),
        ({"cluster_by": ["fake_value", "fake_value_1"]}, "fake_value, fake_value_1"),
    ],
)
def test_cluster_by(config, expected):
    model = make_model(config)
    assert cluster_by(model) == expected


@pytest.mark.parametrize(
    "config,expected",
    [
        ({"partition_by": "fake_value"}, "fake_value"),
        ({"partition_by": ["fake_value", "fake_value_1"]}, ["fake_value", "fake_value_1"]),
    ],
)
def test_partition_by(config, expected):
    model = make_model(config)
    assert partition_by(model) == expected


def test_catalog_name():
    model = make_model({"catalog_name": "my_catalog"})
    assert catalog_name(model).lower() == "my_catalog"


def test_catalog_name_with_non_model_config():
    exported_config = ExportConfig(export_as=ExportDestinationType.TABLE)
    model = make_model(exported_config)
    assert catalog_name(model) == None



# <temporary>_<transient>_table
TRANSIENT_TRUE_TABLE = """
import pandas

def model(dbt, session):
    dbt.config(transient=True)
    return pandas.DataFrame([[1,2]] * 10, columns=['test', 'test2'])
"""


TRANSIENT_FALSE_TABLE = """
import pandas

def model(dbt, session):
    dbt.config(transient=False)
    return pandas.DataFrame([[1,2]] * 10, columns=['test', 'test2'])
"""


TRANSIENT_NONE_TABLE = """
import pandas

def model(dbt, session):
    dbt.config(transient=None)
    return pandas.DataFrame([[1,2]] * 10, columns=['test', 'test2'])
"""


TRANSIENT_UNSET_TABLE = """
import pandas

def model(dbt, session):
    return pandas.DataFrame([[1,2]] * 10, columns=['test', 'test2'])
"""


MACRO__DESCRIBE_TABLES = """
{% macro snowflake__test__describe_tables() %}
    {%- set _sql -%}
        show tables;
        select "name", "kind"
        from table(result_scan(last_query_id()))
    {%- endset %}
    {% set _table = run_query(_sql) %}

    {% do return(_table) %}
{% endmacro %}
"""



import pytest

from dbt.tests.util import run_dbt

from tests.functional.relation_tests.dynamic_table_tests import models
from tests.functional.utils import update_model


class TestCaseSensitiveSupport:
    """
    This test addresses https://github.com/dbt-labs/dbt-adapters/issues/917.

    Setting QUOTED_IDENTIFIERS_IGNORE_CASE = TRUE overrides the quoting behavior in the metadata queries.
    This query:
        select
            "name",
            "schema_name",
            "database_name",
            "text",
            "target_lag",
            "warehouse",
            "refresh_mode"
        from table(result_scan(last_query_id()))
    returns these field names:
        NAME, SCHEMA_NAME, DATABASE_NAME, TEXT, TARGET_LAG, WAREHOUSE, REFRESH_MODE
    when QUOTED_IDENTIFIERS_IGNORE_CASE = TRUE

    Relevant resources:
    https://docs.snowflake.com/en/sql-reference/parameters#quoted-identifiers-ignore-case
    https://docs.snowflake.com/en/sql-reference/identifiers-syntax#controlling-case-using-the-quoted-identifiers-ignore-case-parameter
    """

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        yield {"my_seed.csv": models.SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        yield {"my_dynamic_table.sql": models.DYNAMIC_TABLE}

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "on_configuration_change": "apply",
                "pre-hook": "alter session set quoted_identifiers_ignore_case = true",
                "post-hook": "alter session unset quoted_identifiers_ignore_case",
            }
        }

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        run_dbt(["seed"])
        yield
        project.run_sql(f"drop schema if exists {project.test_schema} cascade")
        project.run_sql(f"alter session unset quoted_identifiers_ignore_case")

    def test_changes_are_applied(self, project):

        # run dbt once and update the table so that the second run needs to query the metadata
        run_dbt(["run"])
        update_model(project, "my_dynamic_table", models.DYNAMIC_TABLE_ALTER)

        # run it a second time to trigger the metadata query
        run_dbt(["run"])

        # a successful run is a pass
        assert True



from dbt.adapters.snowflake.adapter_response import SnowflakeAdapterResponse
from dbt.adapters.snowflake.column import SnowflakeColumn
from dbt.adapters.snowflake.connections import SnowflakeConnectionManager
from dbt.adapters.snowflake.connections import SnowflakeCredentials
from dbt.adapters.snowflake.relation import SnowflakeRelation
from dbt.adapters.snowflake.impl import SnowflakeAdapter

from dbt.adapters.base import AdapterPlugin
from dbt.include import snowflake

Plugin = AdapterPlugin(
    adapter=SnowflakeAdapter,  # type:ignore
    credentials=SnowflakeCredentials,
    include_path=snowflake.PACKAGE_PATH,
)



import pytest
import os

# Import the fuctional fixtures as a plugin
# Note: fixtures with session scope need to be local

pytest_plugins = ["dbt.tests.fixtures.project"]


# The profile dictionary, used to write out profiles.yml
@pytest.fixture(scope="class")
def dbt_profile_target():
    return {
        "type": "snowflake",
        "threads": 4,
        "account": os.getenv("SNOWFLAKE_TEST_ACCOUNT"),
        "user": os.getenv("SNOWFLAKE_TEST_USER"),
        "password": os.getenv("SNOWFLAKE_TEST_PASSWORD"),
        "database": os.getenv("SNOWFLAKE_TEST_DATABASE"),
        "warehouse": os.getenv("SNOWFLAKE_TEST_WAREHOUSE"),
    }



from dataclasses import dataclass
from typing import Optional, Dict, Any, TYPE_CHECKING, Union

from dbt.adapters.relation_configs import RelationConfigChange, RelationResults
from dbt.adapters.contracts.relation import RelationConfig
from dbt.adapters.contracts.relation import ComponentName
from dbt_common.dataclass_schema import StrEnum  # doesn't exist in standard library until py3.11
from typing_extensions import Self

from dbt.adapters.snowflake.parse_model import cluster_by
from dbt.adapters.snowflake.relation_configs.base import SnowflakeRelationConfigBase

if TYPE_CHECKING:
    import agate


class RefreshMode(StrEnum):
    AUTO = "AUTO"
    FULL = "FULL"
    INCREMENTAL = "INCREMENTAL"

    @classmethod
    def default(cls) -> Self:
        return cls("AUTO")


class Initialize(StrEnum):
    ON_CREATE = "ON_CREATE"
    ON_SCHEDULE = "ON_SCHEDULE"

    @classmethod
    def default(cls) -> Self:
        return cls("ON_CREATE")


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class SnowflakeDynamicTableConfig(SnowflakeRelationConfigBase):
    """
    This config follow the specs found here:
    https://docs.snowflake.com/en/sql-reference/sql/create-dynamic-table

    The following parameters are configurable by dbt:
    - name: name of the dynamic table
    - query: the query behind the table
    - target_lag: the maximum amount of time that the dynamic table’s content should lag behind updates to the base tables
    - snowflake_warehouse: the name of the warehouse that provides the compute resources for refreshing the dynamic table
    - refresh_mode: specifies the refresh type for the dynamic table
    - initialize: specifies the behavior of the initial refresh of the dynamic table
    - cluster_by: specifies the columns to cluster on

    There are currently no non-configurable parameters.
    """

    name: str
    schema_name: str
    database_name: str
    query: str
    target_lag: str
    snowflake_warehouse: str
    refresh_mode: Optional[RefreshMode] = RefreshMode.default()
    initialize: Optional[Initialize] = Initialize.default()
    row_access_policy: Optional[str] = None
    table_tag: Optional[str] = None
    cluster_by: Optional[Union[str, list[str]]] = None

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> Self:
        kwargs_dict = {
            "name": cls._render_part(
                ComponentName.Identifier, config_dict.get("name")  # type:ignore
            ),
            "schema_name": cls._render_part(
                ComponentName.Schema, config_dict.get("schema_name")  # type:ignore
            ),
            "database_name": cls._render_part(
                ComponentName.Database, config_dict.get("database_name")  # type:ignore
            ),
            "query": config_dict.get("query"),
            "target_lag": config_dict.get("target_lag"),
            "snowflake_warehouse": config_dict.get("snowflake_warehouse"),
            "refresh_mode": config_dict.get("refresh_mode"),
            "initialize": config_dict.get("initialize"),
            "row_access_policy": config_dict.get("row_access_policy"),
            "table_tag": config_dict.get("table_tag"),
            "cluster_by": config_dict.get("cluster_by"),
        }

        return super().from_dict(kwargs_dict)  # type:ignore

    @classmethod
    def parse_relation_config(cls, relation_config: RelationConfig) -> Dict[str, Any]:
        config_dict = {
            "name": relation_config.identifier,
            "schema_name": relation_config.schema,
            "database_name": relation_config.database,
            "query": relation_config.compiled_code,
            "target_lag": relation_config.config.extra.get("target_lag"),  # type:ignore
            "snowflake_warehouse": relation_config.config.extra.get(  # type:ignore
                "snowflake_warehouse"
            ),
            "row_access_policy": relation_config.config.extra.get(  # type:ignore
                "row_access_policy"
            ),
            "table_tag": relation_config.config.extra.get("table_tag"),  # type:ignore
            "cluster_by": cluster_by(relation_config),
        }

        if refresh_mode := relation_config.config.extra.get("refresh_mode"):  # type:ignore
            config_dict["refresh_mode"] = refresh_mode.upper()

        if initialize := relation_config.config.extra.get("initialize"):  # type:ignore
            config_dict["initialize"] = initialize.upper()

        return config_dict

    @classmethod
    def parse_relation_results(cls, relation_results: RelationResults) -> Dict[str, Any]:
        dynamic_table: "agate.Row" = relation_results["dynamic_table"].rows[0]

        config_dict = {
            "name": dynamic_table.get("name"),
            "schema_name": dynamic_table.get("schema_name"),
            "database_name": dynamic_table.get("database_name"),
            "query": dynamic_table.get("text"),
            "target_lag": dynamic_table.get("target_lag"),
            "snowflake_warehouse": dynamic_table.get("warehouse"),
            "refresh_mode": dynamic_table.get("refresh_mode"),
            "row_access_policy": dynamic_table.get("row_access_policy"),
            "table_tag": dynamic_table.get("table_tag"),
            "cluster_by": dynamic_table.get("cluster_by"),
            # we don't get initialize since that's a one-time scheduler attribute, not a DT attribute
        }

        return config_dict


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class SnowflakeDynamicTableTargetLagConfigChange(RelationConfigChange):
    context: Optional[str] = None

    @property
    def requires_full_refresh(self) -> bool:
        return False


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class SnowflakeDynamicTableWarehouseConfigChange(RelationConfigChange):
    context: Optional[str] = None

    @property
    def requires_full_refresh(self) -> bool:
        return False


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class SnowflakeDynamicTableRefreshModeConfigChange(RelationConfigChange):
    context: Optional[str] = None

    @property
    def requires_full_refresh(self) -> bool:
        return True


@dataclass
class SnowflakeDynamicTableConfigChangeset:
    target_lag: Optional[SnowflakeDynamicTableTargetLagConfigChange] = None
    snowflake_warehouse: Optional[SnowflakeDynamicTableWarehouseConfigChange] = None
    refresh_mode: Optional[SnowflakeDynamicTableRefreshModeConfigChange] = None

    @property
    def requires_full_refresh(self) -> bool:
        return any(
            [
                self.target_lag.requires_full_refresh if self.target_lag else False,
                (
                    self.snowflake_warehouse.requires_full_refresh
                    if self.snowflake_warehouse
                    else False
                ),
                self.refresh_mode.requires_full_refresh if self.refresh_mode else False,
            ]
        )

    @property
    def has_changes(self) -> bool:
        return any([self.target_lag, self.snowflake_warehouse, self.refresh_mode])



import pytest

from dbt.tests.util import assert_message_in_logs, run_dbt, run_dbt_and_capture

from tests.functional.relation_tests.dynamic_table_tests import models
from tests.functional.utils import query_relation_type


class TestBasic:

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {"my_seed.csv": models.SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        yield {
            "my_dynamic_table.sql": models.DYNAMIC_TABLE,
            "my_dynamic_table_downstream.sql": models.DYNAMIC_TABLE_DOWNSTREAM,
            "my_dynamic_iceberg_table.sql": models.DYNAMIC_ICEBERG_TABLE,
        }

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])

    def test_dynamic_table_full_refresh(self, project):
        run_dbt(["run", "--full-refresh"])
        assert query_relation_type(project, "my_dynamic_table") == "dynamic_table"
        assert query_relation_type(project, "my_dynamic_table_downstream") == "dynamic_table"
        assert query_relation_type(project, "my_dynamic_iceberg_table") == "dynamic_table"


class TestAutoConfigDoesntFullRefresh:
    """
    AUTO refresh_strategy will be compared accurately with both INCREMENTAL and FULL.
    https://github.com/dbt-labs/dbt-snowflake/issues/1267
    """

    DT_NAME = "my_dynamic_table"

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {"my_seed.csv": models.SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        yield {
            f"explicit_{self.DT_NAME}.sql": models.EXPLICIT_AUTO_DYNAMIC_TABLE,
            f"implicit_{self.DT_NAME}.sql": models.IMPLICIT_AUTO_DYNAMIC_TABLE,
        }

    @pytest.mark.parametrize("test_dt", [f"explicit_{DT_NAME}", f"implicit_{DT_NAME}"])
    def test_auto_config_doesnt_full_refresh(self, project, test_dt):
        model_qualified_name = f"{project.database}.{project.test_schema}.{test_dt}"

        run_dbt(["seed"])
        _, logs = run_dbt_and_capture(["--debug", "run", "--select", f"{test_dt}.sql"])
        assert_message_in_logs(f"create dynamic table {model_qualified_name}", logs)
        assert_message_in_logs("refresh_mode = AUTO", logs)

        _, logs = run_dbt_and_capture(["--debug", "run", "--select", f"{test_dt}.sql"])

        assert_message_in_logs(f"create dynamic table {model_qualified_name}", logs, False)
        assert_message_in_logs(
            f"create or replace dynamic table {model_qualified_name}", logs, False
        )
        assert_message_in_logs("refresh_mode = AUTO", logs, False)
        assert_message_in_logs(
            f"No configuration changes were identified on: `{model_qualified_name}`. Continuing.",
            logs,
        )



import pytest
from dbt.tests.util import run_dbt


TABLE_BASE_SQL = """
-- models/my_model.sql
{{ config(schema = '1_contains_special*character$') }}
select 1 as id
"""


class TestSpecialCharactersInSchema:

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"quoting": {"schema": True}}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": TABLE_BASE_SQL,
        }

    def test_schema_with_special_chars(self, project):
        run_dbt(["run", "-s", "my_model"])



from dataclasses import dataclass
from typing import Optional

from dbt.adapters.contracts.connection import AdapterResponse


@dataclass
class SnowflakeAdapterResponse(AdapterResponse):
    """Extended AdapterResponse for Snowflake that includes DML statistics.

    These stats are populated from SnowflakeCursor.stats (available in snowflake-connector-python >= 4.2.0)
    and provide granular information about DML operations like CTAS, INSERT, UPDATE, DELETE.
    """

    rows_inserted: Optional[int] = None
    rows_deleted: Optional[int] = None
    rows_updated: Optional[int] = None
    rows_duplicates: Optional[int] = None



import pytest
from dbt.tests.util import run_dbt, write_file


_MODEL_ICEBERG_BASE = """
{{
  config(
    materialized="incremental",
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    on_schema_change="append_new_columns"
  )
}}

select 1 as id,
cast('John' as varchar) as first_name
"""

_MODEL_ICEBERG_ADDED_COLUMN = """
{{
  config(
    materialized="incremental",
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    on_schema_change="append_new_columns"
  )
}}

select 1 as id,
cast('John' as varchar) as first_name,
cast('Smith' as varchar) as last_name
"""

_MODEL_ICEBERG_ADDED_STRING_COLUMN = """
{{
  config(
    materialized="incremental",
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    on_schema_change="append_new_columns"
  )
}}

select 1 as id,
cast('John' as varchar) as first_name,
cast('Smith' as string) as last_name
"""

_MODEL_ICEBERG_ADDED_SIZED_VARCHAR_COLUMN = """
{{
  config(
    materialized="incremental",
    table_format="iceberg",
    external_volume="ICEBERG_SANDBOX",
    catalog="SNOWFLAKE",
    on_schema_change="append_new_columns"
  )
}}

select 1 as id,
cast('John' as varchar) as first_name,
cast('Smith' as varchar(134217728)) as last_name
"""


class TestIcebergSchemaChange:
    """
    Test schema changes with Iceberg tables to ensure VARCHAR columns work correctly.

    This tests the fix for the bug where adding VARCHAR columns to Iceberg tables
    fails because dbt generates VARCHAR(16777216) which is not supported by Snowflake
    Iceberg tables. The fix should use STRING instead for Iceberg tables.
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_iceberg_base.sql": _MODEL_ICEBERG_BASE,
        }

    def test_iceberg_varchar_column_addition(self, project):
        """Test that adding VARCHAR columns to Iceberg tables works correctly."""

        # First, create the initial table
        run_dbt(["run", "--select", "test_iceberg_base"])

        # Verify the table was created successfully
        results = run_dbt(["run", "--select", "test_iceberg_base"])
        assert len(results) == 1

        # Now add a VARCHAR column by updating the model
        write_file(_MODEL_ICEBERG_ADDED_COLUMN, "models", "test_iceberg_base.sql")

        # This should not fail with the varchar size error
        results = run_dbt(["run", "--select", "test_iceberg_base"])
        assert len(results) == 1
        assert results[0].status == "success"

    def test_iceberg_string_column_addition(self, project):
        """Test that adding STRING columns to Iceberg tables works correctly."""

        # First, create the initial table
        run_dbt(["run", "--select", "test_iceberg_base"])

        # Now add a STRING column by updating the model
        write_file(_MODEL_ICEBERG_ADDED_STRING_COLUMN, "models", "test_iceberg_base.sql")

        # This should work fine
        results = run_dbt(["run", "--select", "test_iceberg_base"])
        assert len(results) == 1
        assert results[0].status == "success"

    def test_iceberg_max_varchar_column_addition(self, project):
        """Test that adding VARCHAR with max size to Iceberg tables works correctly."""

        # First, create the initial table
        run_dbt(["run", "--select", "test_iceberg_base"])

        # Now add a VARCHAR column with max size by updating the model
        write_file(_MODEL_ICEBERG_ADDED_SIZED_VARCHAR_COLUMN, "models", "test_iceberg_base.sql")

        # This should work fine
        results = run_dbt(["run", "--select", "test_iceberg_base"])
        assert len(results) == 1
        assert results[0].status == "success"


class TestIcebergSchemaChangeIntegration:

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_iceberg.sql": _MODEL_ICEBERG_BASE,
        }

    def test_reproduce_and_fix_bug(self, project):

        # Step 1: Create the initial incremental iceberg table
        results = run_dbt(["run"])
        assert len(results) == 1
        assert results[0].status == "success"

        # Step 2: Modify the model to add new column (this used to fail)
        write_file(_MODEL_ICEBERG_ADDED_COLUMN, "models", "test_iceberg.sql")

        # Step 3: Run dbt build again - this should now work with our fix
        results = run_dbt(["run"])
        assert len(results) == 1
        assert results[0].status == "success"



import pytest
import shutil
import os
from copy import deepcopy
from dbt.tests.util import run_dbt
from dbt.tests.adapter.dbt_clone.test_dbt_clone import (
    BaseClonePossible,
    BaseCloneSameSourceAndTarget,
)


class TestSnowflakeClonePossible(BaseClonePossible):
    @pytest.fixture(autouse=True)
    def clean_up(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database, schema=f"{project.test_schema}_SEEDS"
            )
            project.adapter.drop_schema(relation)

            relation = project.adapter.Relation.create(
                database=project.database, schema=project.test_schema
            )
            project.adapter.drop_schema(relation)

    pass


table_model_1_sql = """
    {{ config(
        materialized='table',
        transient=true,
    ) }}

    select 1 as fun
    """


class TestSnowflakeCloneTrainsentTable:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "table_model.sql": table_model_1_sql,
        }

    @pytest.fixture(scope="class")
    def other_schema(self, unique_schema):
        return unique_schema + "_other"

    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target, unique_schema, other_schema):
        outputs = {"default": dbt_profile_target, "otherschema": deepcopy(dbt_profile_target)}
        outputs["default"]["schema"] = unique_schema
        outputs["otherschema"]["schema"] = other_schema
        return {"test": {"outputs": outputs, "target": "default"}}

    def copy_state(self, project_root):
        state_path = os.path.join(project_root, "state")
        if not os.path.exists(state_path):
            os.makedirs(state_path)
        shutil.copyfile(
            f"{project_root}/target/manifest.json", f"{project_root}/state/manifest.json"
        )

    def run_and_save_state(self, project_root, with_snapshot=False):
        results = run_dbt(["run"])
        assert len(results) == 1

        self.copy_state(project_root)

    def test_can_clone_transient_table(self, project, other_schema):
        project.create_test_schema(other_schema)
        self.run_and_save_state(project.project_root)

        clone_args = [
            "clone",
            "--state",
            "state",
            "--target",
            "otherschema",
        ]

        results = run_dbt(clone_args)
        assert len(results) == 1


class TestSnowflakeCloneSameSourceAndTarget(BaseCloneSameSourceAndTarget):
    pass



import textwrap
from dataclasses import dataclass, field
from typing import FrozenSet, Optional, Type, Iterator, Tuple


from dbt.adapters.base.relation import BaseRelation, EventTimeFilter
from dbt.adapters.contracts.relation import ComponentName, RelationConfig
from dbt.adapters.events.types import AdapterEventWarning, AdapterEventDebug
from dbt.adapters.relation_configs import (
    RelationConfigBase,
    RelationConfigChangeAction,
    RelationResults,
)
from dbt.adapters.utils import classproperty
from dbt_common.exceptions import DbtRuntimeError
from dbt_common.events.functions import fire_event, warn_or_error

from dbt.adapters.snowflake import constants
from dbt.adapters.snowflake.relation_configs import (
    RefreshMode,
    SnowflakeDynamicTableConfig,
    SnowflakeDynamicTableConfigChangeset,
    SnowflakeDynamicTableRefreshModeConfigChange,
    SnowflakeDynamicTableTargetLagConfigChange,
    SnowflakeDynamicTableWarehouseConfigChange,
    SnowflakeQuotePolicy,
    SnowflakeRelationType,
)


@dataclass(frozen=True, eq=False, repr=False)
class SnowflakeRelation(BaseRelation):
    type: Optional[SnowflakeRelationType] = None
    table_format: str = constants.INFO_SCHEMA_TABLE_FORMAT
    quote_policy: SnowflakeQuotePolicy = field(default_factory=lambda: SnowflakeQuotePolicy())
    require_alias: bool = False
    relation_configs = {
        SnowflakeRelationType.DynamicTable: SnowflakeDynamicTableConfig,
    }
    renameable_relations: FrozenSet[SnowflakeRelationType] = field(
        default_factory=lambda: frozenset(
            {
                SnowflakeRelationType.Table,  # type: ignore
                SnowflakeRelationType.View,  # type: ignore
                SnowflakeRelationType.DynamicTable,  # type: ignore
            }
        )
    )

    replaceable_relations: FrozenSet[SnowflakeRelationType] = field(
        default_factory=lambda: frozenset(
            {
                SnowflakeRelationType.DynamicTable,  # type: ignore
                SnowflakeRelationType.Table,  # type: ignore
                SnowflakeRelationType.View,  # type: ignore
            }
        )
    )

    @property
    def is_dynamic_table(self) -> bool:
        return self.type == SnowflakeRelationType.DynamicTable

    @property
    def is_materialized_view(self) -> bool:
        return self.type == SnowflakeRelationType.DynamicTable

    @property
    def is_iceberg_format(self) -> bool:
        return self.table_format == constants.ICEBERG_TABLE_FORMAT

    @classproperty
    def DynamicTable(cls) -> str:
        return str(SnowflakeRelationType.DynamicTable)

    @classproperty
    def get_relation_type(cls) -> Type[SnowflakeRelationType]:
        return SnowflakeRelationType

    @classmethod
    def from_config(cls, config: RelationConfig) -> RelationConfigBase:
        relation_type: str = config.config.materialized  # type:ignore

        if relation_config := cls.relation_configs.get(relation_type):
            return relation_config.from_relation_config(config)

        raise DbtRuntimeError(
            f"from_config() is not supported for the provided relation type: {relation_type}"
        )

    @classmethod
    def dynamic_table_config_changeset(
        cls, relation_results: RelationResults, relation_config: RelationConfig
    ) -> Optional[SnowflakeDynamicTableConfigChangeset]:
        existing_dynamic_table = SnowflakeDynamicTableConfig.from_relation_results(
            relation_results
        )
        new_dynamic_table = SnowflakeDynamicTableConfig.from_relation_config(relation_config)

        config_change_collection = SnowflakeDynamicTableConfigChangeset()

        if new_dynamic_table.target_lag != existing_dynamic_table.target_lag:
            config_change_collection.target_lag = SnowflakeDynamicTableTargetLagConfigChange(
                action=RelationConfigChangeAction.alter,  # type:ignore
                context=new_dynamic_table.target_lag,
            )

        if new_dynamic_table.snowflake_warehouse != existing_dynamic_table.snowflake_warehouse:
            config_change_collection.snowflake_warehouse = (
                SnowflakeDynamicTableWarehouseConfigChange(
                    action=RelationConfigChangeAction.alter,  # type:ignore
                    context=new_dynamic_table.snowflake_warehouse,
                )
            )

        if (
            new_dynamic_table.refresh_mode != RefreshMode.AUTO
            and new_dynamic_table.refresh_mode != existing_dynamic_table.refresh_mode
        ):
            config_change_collection.refresh_mode = SnowflakeDynamicTableRefreshModeConfigChange(
                action=RelationConfigChangeAction.create,  # type:ignore
                context=new_dynamic_table.refresh_mode,
            )

        if config_change_collection.has_changes:
            return config_change_collection
        return None

    def as_case_sensitive(self) -> "SnowflakeRelation":
        path_part_map = {}

        for path in ComponentName:  # type:ignore
            if self.include_policy.get_part(path):
                part = self.path.get_part(path)
                if part:
                    if self.quote_policy.get_part(path):
                        path_part_map[path] = part
                    else:
                        path_part_map[path] = part.upper()

        return self.replace_path(**path_part_map)

    @property
    def can_be_renamed(self) -> bool:
        """
        Standard tables and dynamic tables can be renamed, but Snowflake does not support renaming iceberg relations.
        The iceberg standard does support renaming, so this may change in the future.
        """
        return self.type in self.renameable_relations and not self.is_iceberg_format

    def get_ddl_prefix_for_create(self, config: RelationConfig, temporary: bool) -> str:
        """
        This macro renders the appropriate DDL prefix during the create_table_as
        macro. It decides based on mutually exclusive table configuration options:
        - TEMPORARY: Indicates a table that exists only for the duration of the session.
        - ICEBERG: A specific storage format that requires a distinct DDL layout.
        - TRANSIENT: A table similar to a permanent table but without fail-safe.
        Additional Caveats for Iceberg models:
        - transient=true throws a warning because Iceberg does not support transient tables
        - A temporary relation is never an Iceberg relation because Iceberg does not
          support temporary relations.
        """

        transient_explicitly_set_true: bool = config.get("transient", False)  # type:ignore

        # Temporary tables are a Snowflake feature that do not exist in the
        # Iceberg framework. We ignore the Iceberg status of the model.
        if temporary:
            return "temporary"
        elif self.is_iceberg_format:
            # Log a warning that transient=true on an Iceberg relation is ignored.
            if transient_explicitly_set_true:
                warn_or_error(
                    AdapterEventWarning(
                        base_msg=(
                            "Iceberg format relations cannot be transient. Please "
                            "remove either the transient or iceberg config options "
                            f"from {self.path.database}.{self.path.schema}."
                            f"{self.path.identifier}. If left unmodified, dbt will "
                            "ignore 'transient'."
                        )
                    )
                )

            return "iceberg"

        # Always supply transient on table create DDL unless user specifically sets
        # transient to false or unset. Might as well update the object attribute too!
        elif transient_explicitly_set_true or config.get("transient", True):  # type:ignore
            return "transient"
        else:
            return ""

    def get_ddl_prefix_for_alter(self) -> str:
        """All ALTER statements on Iceberg tables require an ICEBERG prefix"""
        if self.is_iceberg_format:
            return "iceberg"
        else:
            return ""

    def _render_event_time_filtered(self, event_time_filter: EventTimeFilter) -> str:
        """
        Returns "" if start and end are both None
        """
        filter = ""
        if event_time_filter.start and event_time_filter.end:
            filter = f"{event_time_filter.field_name} >= to_timestamp_tz('{event_time_filter.start}') and {event_time_filter.field_name} < to_timestamp_tz('{event_time_filter.end}')"
        elif event_time_filter.start:
            filter = (
                f"{event_time_filter.field_name} >= to_timestamp_tz('{event_time_filter.start}')"
            )
        elif event_time_filter.end:
            filter = f"{event_time_filter.field_name} < to_timestamp_tz('{event_time_filter.end}')"

        return filter

    def get_iceberg_ddl_options(self, config: RelationConfig) -> str:
        # If the base_location_root config is supplied, overwrite the default value ("_dbt/")
        base_location: str = (
            f"{config.get('base_location_root', '_dbt')}/{self.schema}/{self.name}"  # type:ignore
        )

        if subpath := config.get("base_location_subpath"):  # type:ignore
            base_location += f"/{subpath}"

        external_volume = config.get("external_volume")  # type:ignore
        iceberg_ddl_predicates: str = f"""
        catalog = 'snowflake'
        base_location = '{base_location}'
        """
        if external_volume := config.get("external_volume"):  # type:ignore
            iceberg_ddl_predicates += f"\nexternal_volume = '{external_volume}'"
        return textwrap.indent(textwrap.dedent(iceberg_ddl_predicates), " " * 10)

    def __drop_conditions(self, old_relation: "SnowflakeRelation") -> Iterator[Tuple[bool, str]]:
        drop_view_message: str = (
            f"Dropping relation {old_relation} because it is a view and target relation {self} "
            f"is of type {self.type}."
        )

        drop_table_for_iceberg_message: str = (
            f"Dropping relation {old_relation} because it is a default format table "
            f"and target relation {self} is an Iceberg format table."
        )

        drop_iceberg_for_table_message: str = (
            f"Dropping relation {old_relation} because it is an Iceberg format table "
            f"and target relation {self} is a default format table."
        )

        # An existing view must be dropped for model to build into a table".
        yield (not old_relation.is_table, drop_view_message)
        # An existing table must be dropped for model to build into an Iceberg table.
        yield (
            old_relation.is_table
            and not old_relation.is_iceberg_format
            and self.is_iceberg_format,
            drop_table_for_iceberg_message,
        )
        # existing Iceberg table must be dropped for model to build into a table.
        yield (
            old_relation.is_table
            and old_relation.is_iceberg_format
            and not self.is_iceberg_format,
            drop_iceberg_for_table_message,
        )

    def needs_to_drop(self, old_relation: Optional["SnowflakeRelation"]) -> bool:
        """
        To convert between Iceberg and non-Iceberg relations, a preemptive drop is
        required.

        drops cause latency, but it should be a relatively infrequent occurrence.

        Some Boolean expression below are logically redundant, but this is done for easier
        readability.
        """

        if old_relation is None:
            return False

        for condition, message in self.__drop_conditions(old_relation):
            if condition:
                fire_event(AdapterEventDebug(base_msg=message))
                return True

        return False



from dataclasses import dataclass
from typing import Optional

from dbt.adapters.catalogs import (
    CatalogIntegration,
    CatalogIntegrationConfig,
)
from dbt.adapters.contracts.relation import RelationConfig

from dbt.adapters.snowflake import constants, parse_model


@dataclass
class InfoSchemaCatalogRelation:
    catalog_type: str = constants.DEFAULT_INFO_SCHEMA_CATALOG.catalog_type
    catalog_name: Optional[str] = constants.DEFAULT_INFO_SCHEMA_CATALOG.name
    table_format: Optional[str] = constants.INFO_SCHEMA_TABLE_FORMAT
    external_volume: Optional[str] = None
    file_format: Optional[str] = None
    cluster_by: Optional[str] = None
    automatic_clustering: Optional[bool] = False
    is_transient: Optional[bool] = False


class InfoSchemaCatalogIntegration(CatalogIntegration):
    catalog_name = constants.DEFAULT_INFO_SCHEMA_CATALOG.name
    catalog_type = constants.DEFAULT_INFO_SCHEMA_CATALOG.catalog_type
    table_format = constants.INFO_SCHEMA_TABLE_FORMAT
    allows_writes = True

    def __init__(self, config: CatalogIntegrationConfig) -> None:
        # we overwrite this because the base provides too much config
        self.name: str = config.name
        self.external_volume: Optional[str] = config.external_volume

    def build_relation(self, model: RelationConfig) -> InfoSchemaCatalogRelation:
        """
        Args:
            model: `config.model` (not `model`) from the jinja context
        """
        return InfoSchemaCatalogRelation(
            cluster_by=parse_model.cluster_by(model),
            automatic_clustering=parse_model.automatic_clustering(model),
            is_transient=parse_model.is_transient(model),
        )



from typing import Any, Dict, Optional

from dbt.tests.util import (
    get_connection,
    get_model_file,
    relation_from_name,
    set_model_file,
)

from dbt.adapters.snowflake.relation_configs import SnowflakeDynamicTableConfig


def query_relation_type(project, name: str) -> Optional[str]:
    relation = relation_from_name(project.adapter, name)
    sql = f"""
        select
            case table_type
                when 'BASE TABLE' then iff(is_dynamic = 'YES', 'dynamic_table', 'table')
                when 'VIEW' then 'view'
                when 'EXTERNAL TABLE' then 'external_table'
            end as relation_type
        from information_schema.tables
        where table_name like '{relation.identifier.upper()}'
        and table_schema like '{relation.schema.upper()}'
        and table_catalog like '{relation.database.upper()}'
    """
    results = project.run_sql(sql, fetch="all")

    assert len(results) > 0, f"Relation {relation} not found"
    assert len(results) == 1, f"Multiple relations found"

    return results[0][0].lower()


def query_row_count(project, name: str) -> int:
    relation = relation_from_name(project.adapter, name)
    sql = f"select count(*) from {relation}"
    return project.run_sql(sql, fetch="one")[0]


def insert_record(project, name: str, record: Dict[str, Any]):
    relation = relation_from_name(project.adapter, name)
    column_names = ", ".join(record.keys())
    values = ", ".join(
        [f"'{value}'" if isinstance(value, str) else f"{value}" for value in record.values()]
    )
    sql = f"insert into {relation} ({column_names}) values ({values})"
    project.run_sql(sql)


def update_model(project, name: str, model: str) -> str:
    relation = relation_from_name(project.adapter, name)
    original_model = get_model_file(project, relation)
    set_model_file(project, relation, model)
    return original_model


def describe_dynamic_table(project, name: str) -> Optional[SnowflakeDynamicTableConfig]:
    macro = "snowflake__describe_dynamic_table"
    dynamic_table = relation_from_name(project.adapter, name)
    kwargs = {"relation": dynamic_table}
    with get_connection(project.adapter):
        results = project.adapter.execute_macro(macro, kwargs=kwargs)

    assert len(results["dynamic_table"].rows) > 0, f"Dynamic table {dynamic_table} not found"
    found = len(results["dynamic_table"].rows)
    names = ", ".join([table.get("name") for table in results["dynamic_table"].rows])
    assert found == 1, f"Multiple dynamic tables found: {names}"

    return SnowflakeDynamicTableConfig.from_relation_results(results)


def refresh_dynamic_table(project, name: str) -> None:
    macro = "snowflake__refresh_dynamic_table"
    dynamic_table = relation_from_name(project.adapter, name)
    kwargs = {"relation": dynamic_table}
    with get_connection(project.adapter):
        project.adapter.execute_macro(macro, kwargs=kwargs)



import dataclasses
from typing import Optional

from dbt_common.record import Record, Recorder


@dataclasses.dataclass
class CursorGetSfqidParams:
    connection_name: str


@dataclasses.dataclass
class CursorGetSfqidResult:
    msg: Optional[str]


@Recorder.register_record_type
class CursorGetSfqidRecord(Record):
    params_cls = CursorGetSfqidParams
    result_cls = CursorGetSfqidResult
    group = "Database"



from dbt.tests.util import run_dbt, check_relations_equal, rm_file, write_file
from dbt.tests.adapter.concurrency.test_concurrency import BaseConcurrency, seeds__update_csv


class TestConncurenncySnowflake(BaseConcurrency):
    def test_conncurrency_snowflake(self, project):
        run_dbt(["seed", "--select", "seed"])
        results = run_dbt(["run"], expect_pass=False)
        assert len(results) == 7
        check_relations_equal(project.adapter, ["SEED", "VIEW_MODEL"])
        check_relations_equal(project.adapter, ["SEED", "DEP"])
        check_relations_equal(project.adapter, ["SEED", "TABLE_A"])
        check_relations_equal(project.adapter, ["SEED", "TABLE_B"])

        rm_file(project.project_root, "seeds", "seed.csv")
        write_file(seeds__update_csv, project.project_root + "/seeds", "seed.csv")
        results = run_dbt(["run"], expect_pass=False)
        assert len(results) == 7
        check_relations_equal(project.adapter, ["SEED", "VIEW_MODEL"])
        check_relations_equal(project.adapter, ["SEED", "DEP"])
        check_relations_equal(project.adapter, ["SEED", "TABLE_A"])
        check_relations_equal(project.adapter, ["SEED", "TABLE_B"])



from dataclasses import dataclass
from typing import Any, Dict, Optional, TYPE_CHECKING
from dbt.adapters.base.relation import Policy
from dbt.adapters.relation_configs import (
    RelationConfigBase,
    RelationResults,
)

from dbt.adapters.contracts.relation import ComponentName, RelationConfig

from dbt.adapters.snowflake.relation_configs.policies import (
    SnowflakeIncludePolicy,
    SnowflakeQuotePolicy,
)

if TYPE_CHECKING:
    # Imported downfile for specific row gathering function.
    import agate


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class SnowflakeRelationConfigBase(RelationConfigBase):
    """
    This base class implements a few boilerplate methods and provides some light structure for Snowflake relations.
    """

    @classmethod
    def include_policy(cls) -> Policy:
        return SnowflakeIncludePolicy()

    @classmethod
    def quote_policy(cls) -> Policy:
        return SnowflakeQuotePolicy()

    @classmethod
    def from_relation_config(cls, relation_config: RelationConfig):
        relation_config_dict = cls.parse_relation_config(relation_config)
        relation = cls.from_dict(relation_config_dict)
        return relation

    @classmethod
    def parse_relation_config(cls, relation_config: RelationConfig) -> Dict:
        raise NotImplementedError(
            "`parse_relation_config()` needs to be implemented on this RelationConfigBase instance"
        )

    @classmethod
    def from_relation_results(cls, relation_results: RelationResults):
        relation_config = cls.parse_relation_results(relation_results)
        relation = cls.from_dict(relation_config)
        return relation

    @classmethod
    def parse_relation_results(cls, relation_results: RelationResults) -> Dict[str, Any]:
        raise NotImplementedError(
            "`parse_relation_results()` needs to be implemented on this RelationConfigBase instance"
        )

    @classmethod
    def _render_part(cls, component: ComponentName, value: Optional[str]) -> Optional[str]:
        if cls.include_policy().get_part(component) and value:
            if cls.quote_policy().get_part(component):
                return f'"{value}"'
            return value.lower()
        return None

    @classmethod
    def _get_first_row(cls, results: "agate.Table") -> "agate.Row":
        try:
            return results.rows[0]
        except IndexError:
            import agate

            return agate.Row(values=set())



from dataclasses import dataclass
from typing import Optional, Union

from dbt.adapters.catalogs import (
    CatalogIntegration,
    CatalogIntegrationConfig,
)
from dbt.adapters.contracts.relation import RelationConfig

from dbt.adapters.snowflake import constants, parse_model
from dbt.adapters.snowflake.constants import SnowflakeIcebergTableRelationParameters

_BOOL_TO_STR_MAP = {
    True: "TRUE",
    False: "FALSE",
    "true": "TRUE",
    "false": "FALSE",
}


def _validate_storage_serialization_policy(policy: str) -> None:
    if policy.upper() not in ["COMPATIBLE", "OPTIMIZED"]:
        raise ValueError(
            f"Invalid storage serialization policy: {policy}. Expected 'COMPATIBLE' or 'OPTIMIZED'."
        )


@dataclass
class BuiltInCatalogRelation:
    """
    Parameters representing configuration of a built-in Snowflake Iceberg table
    Ref: https://docs.snowflake.com/en/sql-reference/sql/create-iceberg-table-snowflake#syntax

    Note that a lot of these parameters are optional, and the defaults for are inherited from the
    schema/database/account. To not add to the complexity of the user configuration, we do not set defaults for these
    parameters.
    """

    base_location: Optional[str]
    catalog_type: str = constants.DEFAULT_BUILT_IN_CATALOG.catalog_type
    catalog_name: Optional[str] = constants.DEFAULT_BUILT_IN_CATALOG.name
    table_format: Optional[str] = constants.ICEBERG_TABLE_FORMAT
    external_volume: Optional[str] = None
    file_format: Optional[str] = None
    cluster_by: Optional[str] = None
    partition_by: Optional[Union[list[str], str]] = None
    automatic_clustering: Optional[bool] = False
    is_transient: Optional[bool] = False
    data_retention_time_in_days: Optional[int] = None
    max_data_extension_time_in_days: Optional[int] = None
    storage_serialization_policy: Optional[str] = None
    change_tracking: Optional[str] = None


class BuiltInCatalogIntegration(CatalogIntegration):
    catalog_name = constants.DEFAULT_BUILT_IN_CATALOG.name
    catalog_type = constants.DEFAULT_BUILT_IN_CATALOG.catalog_type
    table_format = constants.ICEBERG_TABLE_FORMAT
    file_format = None  # no file format for built-in catalogs
    allows_writes = True
    data_retention_time_in_days: Optional[int] = None
    max_data_extension_time_in_days: Optional[int] = None
    storage_serialization_policy = None
    change_tracking = None

    def __init__(self, config: CatalogIntegrationConfig) -> None:
        # we overwrite this because the base provides too much config
        self.name: str = config.name
        self.external_volume: Optional[str] = config.external_volume
        if adapter_properties := config.adapter_properties:
            if storage_serialization_policy := adapter_properties.get(
                SnowflakeIcebergTableRelationParameters.storage_serialization_policy
            ):
                _validate_storage_serialization_policy(storage_serialization_policy)
                self.storage_serialization_policy = storage_serialization_policy

            self.max_data_extension_time_in_days = adapter_properties.get(
                SnowflakeIcebergTableRelationParameters.max_data_extension_time_in_days
            )

            self.change_tracking = adapter_properties.get(
                SnowflakeIcebergTableRelationParameters.change_tracking
            )

            self.data_retention_time_in_days = adapter_properties.get(
                SnowflakeIcebergTableRelationParameters.data_retention_time_in_days
            )

    def build_relation(self, model: RelationConfig) -> BuiltInCatalogRelation:
        """
        Args:
            model: `config.model` (not `model`) from the jinja context
        """

        max_data_extension_time_in_days = (
            parse_model.max_data_extension_time_in_days(model)
            or self.max_data_extension_time_in_days
        )
        data_retention_time_in_days = (
            model.config.get(
                SnowflakeIcebergTableRelationParameters.data_retention_time_in_days,
                self.data_retention_time_in_days,
            )
            if model.config
            else self.data_retention_time_in_days
        )
        storage_serialization_policy = (
            model.config.get(
                SnowflakeIcebergTableRelationParameters.storage_serialization_policy,
                self.storage_serialization_policy,
            )
            if model.config
            else self.storage_serialization_policy
        )

        return BuiltInCatalogRelation(
            base_location=parse_model.base_location(model),
            external_volume=parse_model.external_volume(model) or self.external_volume,
            cluster_by=parse_model.cluster_by(model),
            partition_by=parse_model.partition_by(model),
            automatic_clustering=parse_model.automatic_clustering(model),
            storage_serialization_policy=storage_serialization_policy,
            max_data_extension_time_in_days=max_data_extension_time_in_days,
            change_tracking=self._resolve_change_tracking(model),
            data_retention_time_in_days=data_retention_time_in_days,
        )

    def _resolve_change_tracking(self, model: RelationConfig) -> Optional[str]:
        """
        Resolves the change tracking for the catalog integration.
        If `change_tracking` is set in the model config, it will override the default.
        If 'change_tracking' is not set on either the model or integration, it will return None.
        """
        if (
            model.config
            and (
                change_tracking := model.config.get(
                    SnowflakeIcebergTableRelationParameters.change_tracking, self.change_tracking
                )
            )
            is not None
        ):
            if isinstance(change_tracking, str):
                change_tracking = change_tracking.lower()
            try:
                return _BOOL_TO_STR_MAP[change_tracking]
            except KeyError:
                raise ValueError("Invalid value for change_tracking. Expected 'true' or 'false'.")
        else:
            return None



from dbt.adapters.snowflake.connections import SnowflakeConnectionManager
from dbt.adapters.snowflake.record.cursor.cursor import SnowflakeRecordReplayCursor
from dbt.adapters.snowflake.record.handle import SnowflakeRecordReplayHandle


class MockStats:
    """Mock object that mimics snowflake-connector-python's QueryResultStats."""

    def __init__(
        self,
        num_rows_inserted=100,
        num_rows_deleted=10,
        num_rows_updated=5,
        num_dml_duplicates=2,
    ):
        self.num_rows_inserted = num_rows_inserted
        self.num_rows_deleted = num_rows_deleted
        self.num_rows_updated = num_rows_updated
        self.num_dml_duplicates = num_dml_duplicates


class MockCursor:
    """Mock cursor that mimics snowflake-connector-python's SnowflakeCursor."""

    def __init__(self, stats=None):
        self._stats = stats

    @property
    def rowcount(self) -> int:
        return 42

    @property
    def sqlstate(self) -> str:
        return "00000"

    @property
    def sfqid(self) -> str:
        return "01abc123-0001-abcd-0000-00012345abcd"

    @property
    def stats(self):
        return self._stats

    def execute(self, operation, parameters=None) -> None:
        pass

    @property
    def unexpected_prop(self) -> bool:
        return True

    def unexpected_func(self) -> int:
        return 1


class MockHandle:
    """Mock handle that mimics snowflake-connector-python's SnowflakeConnection."""

    def __init__(self, session_id=123456789):
        self._session_id = session_id

    @property
    def session_id(self) -> int:
        return self._session_id

    def cursor(self):
        return MockCursor()

    @property
    def unexpected_prop(self) -> bool:
        return True

    def unexpected_func(self) -> int:
        return 1


class MockConnection:
    name = "test_connection"


def test_snowflake_record_cursor_sqlstate():
    """Test that the sqlstate property works correctly."""
    recorded_cursor = SnowflakeRecordReplayCursor(MockCursor(), MockConnection())  # type: ignore
    assert recorded_cursor.sqlstate == "00000"


def test_snowflake_record_cursor_sfqid():
    """Test that the sfqid property works correctly."""
    recorded_cursor = SnowflakeRecordReplayCursor(MockCursor(), MockConnection())  # type: ignore
    assert recorded_cursor.sfqid == "01abc123-0001-abcd-0000-00012345abcd"


def test_snowflake_record_cursor_stats():
    """Test that the stats property works correctly."""
    mock_stats = MockStats()
    recorded_cursor = SnowflakeRecordReplayCursor(
        MockCursor(stats=mock_stats), MockConnection()
    )  # type: ignore

    stats = recorded_cursor.stats
    assert stats.num_rows_inserted == 100
    assert stats.num_rows_deleted == 10
    assert stats.num_rows_updated == 5
    assert stats.num_dml_duplicates == 2


def test_snowflake_record_cursor_stats_none():
    """Test that the stats property handles None correctly."""
    recorded_cursor = SnowflakeRecordReplayCursor(
        MockCursor(stats=None), MockConnection()
    )  # type: ignore

    assert recorded_cursor.stats is None


def test_snowflake_record_cursor_inherited_properties():
    """Test that inherited properties from RecordReplayCursor work correctly."""
    recorded_cursor = SnowflakeRecordReplayCursor(MockCursor(), MockConnection())  # type: ignore

    # Test inherited rowcount property
    assert recorded_cursor.rowcount == 42

    # Test inherited execute method
    recorded_cursor.execute("SELECT 1")


def test_snowflake_record_cursor_unexpected_access():
    """Test that unexpected property/method access fires a warning but still works."""
    recorded_cursor = SnowflakeRecordReplayCursor(MockCursor(), MockConnection())  # type: ignore

    events = []
    # Mock event firing
    recorded_cursor._fire_event = events.append

    # Test that an unexpected property works, but fires a warning
    assert recorded_cursor.unexpected_prop is True
    assert len(events) == 1
    assert events[0].__class__.__name__ == "RecordReplayIssue"
    assert "unexpected_prop" in events[0].msg
    events.clear()

    # Test that an unexpected function works, but fires a warning
    assert recorded_cursor.unexpected_func() == 1
    assert len(events) == 1
    assert events[0].__class__.__name__ == "RecordReplayIssue"
    assert "unexpected_func" in events[0].msg


def test_get_response_no_unexpected_access_warnings():
    """Ensure get_response() doesn't trigger any unexpected attribute access warnings.

    This is a regression test. If new cursor attributes are accessed in get_response()
    without being added to SnowflakeRecordReplayCursor, this test will fail.
    """
    events = []

    # Test with stats present
    mock_cursor = MockCursor(stats=MockStats())
    recorded_cursor = SnowflakeRecordReplayCursor(mock_cursor, MockConnection())  # type: ignore
    recorded_cursor._fire_event = events.append

    # Call get_response - this is the actual code path
    response = SnowflakeConnectionManager.get_response(recorded_cursor)

    # Verify no unexpected access warnings were fired
    assert len(events) == 0, (
        f"Unexpected attribute access in get_response(): {[e.msg for e in events]}. "
        "Add the missing attribute(s) to SnowflakeRecordReplayCursor."
    )

    # Verify the response was created successfully
    assert response is not None
    assert response.code == "00000"  # SQL success state from mock cursor


# =============================================================================
# SnowflakeRecordReplayHandle Tests
# =============================================================================


def test_snowflake_record_handle_session_id():
    """Test that the session_id property works correctly."""
    recorded_handle = SnowflakeRecordReplayHandle(MockHandle(), MockConnection())  # type: ignore
    assert recorded_handle.session_id == 123456789


def test_snowflake_record_handle_cursor():
    """Test that cursor() returns a SnowflakeRecordReplayCursor."""
    recorded_handle = SnowflakeRecordReplayHandle(MockHandle(), MockConnection())  # type: ignore
    cursor = recorded_handle.cursor()
    assert isinstance(cursor, SnowflakeRecordReplayCursor)


def test_snowflake_record_handle_unexpected_access():
    """Test that unexpected property/method access fires a warning but still works."""
    recorded_handle = SnowflakeRecordReplayHandle(MockHandle(), MockConnection())  # type: ignore

    events = []
    # Mock event firing
    recorded_handle._fire_event = events.append

    # Test that an unexpected property works, but fires a warning
    assert recorded_handle.unexpected_prop is True
    assert len(events) == 1
    assert events[0].__class__.__name__ == "RecordReplayIssue"
    assert "unexpected_prop" in events[0].msg
    events.clear()

    # Test that an unexpected function works, but fires a warning
    assert recorded_handle.unexpected_func() == 1
    assert len(events) == 1
    assert events[0].__class__.__name__ == "RecordReplayIssue"
    assert "unexpected_func" in events[0].msg


def test_handle_session_id_no_unexpected_access_warnings():
    """Ensure accessing session_id doesn't trigger unexpected attribute access warnings.

    This is a regression test. If new handle attributes are accessed in cancel()
    or elsewhere without being added to SnowflakeRecordReplayHandle, this test will fail.
    """
    events = []

    recorded_handle = SnowflakeRecordReplayHandle(MockHandle(), MockConnection())  # type: ignore
    recorded_handle._fire_event = events.append

    # Access session_id - this is used in SnowflakeConnectionManager.cancel()
    _ = recorded_handle.session_id

    # Verify no unexpected access warnings were fired
    assert len(events) == 0, (
        f"Unexpected attribute access on handle: {[e.msg for e in events]}. "
        "Add the missing attribute(s) to SnowflakeRecordReplayHandle."
    )



_MODEL_SQL = """
select
    1::smallint as smallint_col,
    2::int as int_col,
    3::bigint as bigint_col,
    4::integer as integer_col,
    5::tinyint as tinyint_col,
    6::byteint as byteint_col,
    7.0::float as float_col,
    8.0::float4 as float4_col,
    9.0::float8 as float8_col,
    10.0::double as double_col,
    11.0::double precision as double_p_col,
    12.0::real as real_col,
    13.0::numeric as numeric_col,
    14.0::decimal as decimal_col,
    15.0::number as number_col,
    '16'::text as text_col,
    '17'::varchar(20) as varchar_col
"""

_SCHEMA_YML = """
version: 2
models:
  - name: model
    data_tests:
      - is_type:
          column_map:
            smallint_col: ['numeric', 'number', 'not string', 'not float', 'not integer']
            int_col: ['numeric', 'number', 'not string', 'not float', 'not integer']
            bigint_col: ['numeric', 'number', 'not string', 'not float', 'not integer']
            integer_col: ['numeric', 'number', 'not string', 'not float', 'not integer']
            tinyint_col: ['numeric', 'number', 'not string', 'not float', 'not integer']
            byteint_col: ['numeric', 'number', 'not string', 'not float', 'not integer']
            float_col: ['float', 'number', 'not string', 'not integer', 'not numeric']
            float4_col: ['float', 'number', 'not string', 'not integer', 'not numeric']
            float8_col: ['float', 'number', 'not string', 'not integer', 'not numeric']
            double_col: ['float', 'number', 'not string', 'not integer', 'not numeric']
            double_p_col: ['float', 'number', 'not string', 'not integer', 'not numeric']
            real_col: ['float', 'number', 'not string', 'not integer', 'not numeric']
            numeric_col: ['numeric', 'number', 'not string', 'not float', 'not integer']
            decimal_col: ['numeric', 'number', 'not string', 'not float', 'not integer']
            number_col: ['numeric', 'number', 'not string', 'not float', 'not integer']
            text_col: ['string', 'not number']
            varchar_col: ['string', 'not number']
"""



import re
import pytest
from dbt.tests.adapter.catalog_integrations.test_catalog_integration import (
    BaseCatalogIntegrationValidation,
)
from dbt.tests.util import run_dbt


def get_cleaned_model_ddl_from_file(file_name: str) -> str:
    with open(f"target/run/test/models/{file_name}", "r") as ddl_file:
        return re.sub(r"\s+", " ", ddl_file.read())


MODEL__BASIC_ICEBERG_TABLE = """
                            {{ config(materialized='table', catalog_name='basic_iceberg_catalog') }}
                            select 1 as id
                            """

MODEL__ICEBERG_TABLE_W_CONFIGS = """
                            {{ config(materialized='table',
                                catalog_name='basic_iceberg_catalog',
                                data_retention_time_in_days=1,
                                change_tracking=False,
                                max_data_extension_time_in_days=30,
                                storage_serialization_policy='COMPATIBLE')
                                }}
                            select 1 as id
                            """


class TestSnowflakeBuiltInCatalogIntegration(BaseCatalogIntegrationValidation):

    @pytest.fixture(scope="class")
    def catalogs(self):
        return {
            "catalogs": [
                {
                    "name": "basic_iceberg_catalog",
                    "active_write_integration": "basic_iceberg_catalog_integration",
                    "write_integrations": [
                        {
                            "name": "basic_iceberg_catalog_integration",
                            "catalog_type": "BUILT_IN",
                            "table_format": "iceberg",
                            "external_volume": f"s3_iceberg_snow",
                            "adapter_properties": {
                                "storage_serialization_policy": "OPTIMIZED",
                                "max_data_extension_time_in_days": 60,
                                "data_retention_time_in_days": 0,
                                "change_tracking": True,
                            },
                        }
                    ],
                },
            ]
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "basic_iceberg_table.sql": MODEL__BASIC_ICEBERG_TABLE,
            "iceberg_table_with_configs.sql": MODEL__ICEBERG_TABLE_W_CONFIGS,
        }

    def test_basic_iceberg_catalog_integration(self, project):
        run_dbt(["run"])
        iceberg_sql = get_cleaned_model_ddl_from_file("basic_iceberg_table.sql")
        assert "storage_serialization_policy = 'OPTIMIZED'" in iceberg_sql
        assert "max_data_extension_time_in_days = 60" in iceberg_sql
        assert "change_tracking = TRUE" in iceberg_sql
        assert "data_retention_time_in_days = 0" in iceberg_sql
        iceberg_table_with_configs_sql = get_cleaned_model_ddl_from_file(
            "iceberg_table_with_configs.sql"
        )
        assert "storage_serialization_policy = 'COMPATIBLE'" in iceberg_table_with_configs_sql
        assert "max_data_extension_time_in_days = 30" in iceberg_table_with_configs_sql
        assert "change_tracking = FALSE" in iceberg_table_with_configs_sql
        assert "data_retention_time_in_days = 1" in iceberg_table_with_configs_sql



import pytest
from dbt.tests.adapter.incremental.test_incremental_predicates import BaseIncrementalPredicates


class TestIncrementalPredicatesDeleteInsertSnowflake(BaseIncrementalPredicates):
    pass


class TestPredicatesDeleteInsertSnowflake(BaseIncrementalPredicates):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"+predicates": ["id != 2"], "+incremental_strategy": "delete+insert"}}


class TestIncrementalPredicatesMergeSnowflake(BaseIncrementalPredicates):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+incremental_predicates": ["dbt_internal_dest.id != 2"],
                "+incremental_strategy": "merge",
            }
        }


class TestPredicatesMergeSnowflake(BaseIncrementalPredicates):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+predicates": ["dbt_internal_dest.id != 2"],
                "+incremental_strategy": "merge",
            }
        }



import argparse
import json
import secrets
import textwrap
from base64 import b64encode

import requests
from werkzeug.utils import redirect
from werkzeug.middleware.dispatcher import DispatcherMiddleware
from werkzeug.wrappers import Request, Response
from werkzeug.serving import run_simple
from urllib.parse import urlencode


def _make_rfp_claim_value():
    # from https://tools.ietf.org/id/draft-bradley-oauth-jwt-encoded-state-08.html#rfc.section.4  # noqa
    # we can do whatever we want really, so just token.urlsafe?
    return secrets.token_urlsafe(112)


def _make_response(client_id, client_secret, refresh_token):
    return Response(
        textwrap.dedent(
            f'''\
        SNOWFLAKE_TEST_OAUTH_REFRESH_TOKEN="{refresh_token}"
        SNOWFLAKE_TEST_OAUTH_CLIENT_ID="{client_id}"
        SNOWFLAKE_TEST_OAUTH_CLIENT_SECRET="{client_secret}"'''
        )
    )


class TokenManager:
    def __init__(self, account_name, client_id, client_secret):
        self.account_name = account_name
        self.client_id = client_id
        self.client_secret = client_secret
        self.token = None
        self.rfp_claim = _make_rfp_claim_value()
        self.port = 8080

    @property
    def account_url(self):
        return f"https://{self.account_name}.snowflakecomputing.com"

    @property
    def auth_url(self):
        return f"{self.account_url}/oauth/authorize"

    @property
    def token_url(self):
        return f"{self.account_url}/oauth/token-request"

    @property
    def redirect_uri(self):
        return f"http://localhost:{self.port}"

    @property
    def headers(self):
        auth = f"{self.client_id}:{self.client_secret}".encode("ascii")
        encoded_auth = b64encode(auth).decode("ascii")
        return {
            "Authorization": f"Basic {encoded_auth}",
            "Content-type": "application/x-www-form-urlencoded; charset=utf-8",
        }

    def _code_to_token(self, code):
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self.redirect_uri,
        }
        # data = urlencode(data)
        resp = requests.post(
            url=self.token_url,
            headers=self.headers,
            data=data,
        )
        try:
            refresh_token = resp.json()["refresh_token"]
        except KeyError:
            print(resp.json())
            raise
        return refresh_token

    @Request.application
    def auth(self, request):
        code = request.args.get("code")
        if code:
            # we got 303'ed here with a code
            state_received = request.args.get("state")
            if state_received != self.rfp_claim:
                return Response("Invalid RFP claim: MITM?", status=401)
            refresh_token = self._code_to_token(code)
            return _make_response(
                self.client_id,
                self.client_secret,
                refresh_token,
            )
        else:
            return redirect("/login")

    @Request.application
    def login(self, request):
        # take the auth URL and add the query string to it
        query = {
            "response_type": "code",
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "state": self.rfp_claim,
        }
        query = urlencode(query)
        return redirect(f"{self.auth_url}?{query}")


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("account_name", help="The account name")
    parser.add_argument("json_blob", help="The json auth blob")

    return parser.parse_args()


def main():
    args = parse_args()
    data = json.loads(args.json_blob)
    client_id = data["OAUTH_CLIENT_ID"]
    client_secret = data["OAUTH_CLIENT_SECRET"]
    token_manager = TokenManager(
        account_name=args.account_name,
        client_id=client_id,
        client_secret=client_secret,
    )
    app = DispatcherMiddleware(
        token_manager.auth,
        {
            "/login": token_manager.login,
        },
    )

    run_simple("localhost", token_manager.port, app)


if __name__ == "__main__":
    main()



import base64
import datetime
import os
import sys

if sys.version_info < (3, 9):
    from functools import lru_cache

    cache = lru_cache(maxsize=None)
else:
    from functools import cache

import pytz
import re
from contextlib import contextmanager
from dataclasses import dataclass
from io import StringIO
from time import sleep

from typing import Optional, Tuple, Union, Any, List, Iterable, TYPE_CHECKING, Dict

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
import requests
import snowflake.connector
import snowflake.connector.constants
import snowflake.connector.errors
from snowflake.connector.errors import (
    Error,
    DatabaseError,
    InternalError,
    InternalServerError,
    ServiceUnavailableError,
    GatewayTimeoutError,
    RequestTimeoutError,
    BadGatewayError,
    OtherHTTPRetryableError,
    BindUploadError,
)

from dbt_common.exceptions import (
    DbtInternalError,
    DbtRuntimeError,
    DbtConfigError,
)
from dbt_common.exceptions import DbtDatabaseError
from dbt_common.record import get_record_mode_from_env, RecorderMode
from dbt.adapters.exceptions.connection import FailedToConnectError
from dbt.adapters.contracts.connection import AdapterResponse, Connection, Credentials
from dbt.adapters.snowflake.adapter_response import SnowflakeAdapterResponse
from dbt.adapters.sql import SQLConnectionManager
from dbt.adapters.events.logging import AdapterLogger
from dbt_common.events.functions import warn_or_error
from dbt.adapters.events.types import AdapterEventWarning, AdapterEventError
from dbt_common.ui import line_wrap_message, warning_tag
from dbt.adapters.snowflake.record import SnowflakeRecordReplayHandle

from dbt.adapters.snowflake.auth import private_key_from_file, private_key_from_string
from dbt.adapters.snowflake.query_headers import SnowflakeMacroQueryStringSetter

if TYPE_CHECKING:
    import agate


logger = AdapterLogger("Snowflake")


def setup_snowflake_logging(level: str):
    for logger_name in ["snowflake.connector", "botocore", "boto3"]:
        logger.debug(f"Setting {logger_name} to {level} (file logging only)")
        logger.set_adapter_dependency_log_level(logger_name, level)


if snowflake_level := os.getenv("DBT_SNOWFLAKE_CONNECTOR_DEBUG_LOGGING"):
    if snowflake_level.upper() in ["INFO", "DEBUG", "ERROR"]:
        setup_snowflake_logging(snowflake_level.upper())
    else:
        setup_snowflake_logging("DEBUG")
else:
    setup_snowflake_logging("ERROR")


_TOKEN_REQUEST_URL = "https://{}.snowflakecomputing.com/oauth/token-request"

ERROR_REDACTION_PATTERNS = {
    re.compile(r"Row Values: \[(.|\n)*\]"): "Row Values: [redacted]",
    re.compile(r"Duplicate field key '(.|\n)*'"): "Duplicate field key '[redacted]'",
}


@cache
def snowflake_private_key(private_key: RSAPrivateKey) -> bytes:
    return private_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )


@dataclass
class SnowflakeCredentials(Credentials):
    account: str
    user: Optional[str] = None
    warehouse: Optional[str] = None
    role: Optional[str] = None
    password: Optional[str] = None
    authenticator: Optional[str] = None
    private_key: Optional[str] = None
    private_key_path: Optional[str] = None
    private_key_passphrase: Optional[str] = None
    token: Optional[str] = None
    oauth_client_id: Optional[str] = None
    oauth_client_secret: Optional[str] = None
    query_tag: Optional[str] = None
    client_session_keep_alive: bool = False
    host: Optional[str] = None
    port: Optional[int] = None
    proxy_host: Optional[str] = None
    proxy_port: Optional[int] = None
    protocol: Optional[str] = None
    connect_retries: int = 1
    connect_timeout: Optional[int] = None
    retry_on_database_errors: bool = False
    retry_all: bool = False
    insecure_mode: Optional[bool] = False
    # this needs to default to `None` so that we can tell if the user set it; see `__post_init__()`
    reuse_connections: Optional[bool] = None
    s3_stage_vpce_dns_name: Optional[str] = None
    # Setting this to 0.0 will disable platform detection which adds query latency
    # this should only be set to a non-zero value if you are using WIF authentication
    platform_detection_timeout_seconds: float = 0.0

    def __post_init__(self):
        if self.authenticator != "oauth" and (self.oauth_client_secret or self.oauth_client_id):
            # the user probably forgot to set 'authenticator' like I keep doing
            warn_or_error(
                AdapterEventWarning(
                    base_msg="Authenticator is not set to oauth, but an oauth-only parameter is set! Did you mean to set authenticator: oauth?"
                )
            )

        if self.authenticator not in ["oauth", "jwt"]:
            if self.token:
                warn_or_error(
                    AdapterEventWarning(
                        base_msg=(
                            "The token parameter was set, but the authenticator was "
                            "not set to 'oauth' or 'jwt'."
                        )
                    )
                )

            if not self.user:
                # The user attribute is only optional if 'authenticator' is 'jwt' or 'oauth'
                warn_or_error(
                    AdapterEventError(base_msg="Invalid profile: 'user' is a required property.")
                )

        self.account, sub_count = re.subn("_", "-", self.account)
        if sub_count:
            logger.debug(
                "Replaced underscores (_) with hyphens (-) in Snowflake account name to form a valid account URL."
            )

        # only default `reuse_connections` to `True` if the user has not turned on `client_session_keep_alive`
        # having both of these set to `True` could lead to hanging open connections, so it should be opt-in behavior
        if self.client_session_keep_alive is False and self.reuse_connections is None:
            self.reuse_connections = True

    @property
    def type(self):
        return "snowflake"

    @property
    def unique_field(self):
        return self.account

    # the results show up in the output of dbt debug runs, for more see..
    # https://docs.getdbt.com/guides/dbt-ecosystem/adapter-development/3-building-a-new-adapter#editing-the-connection-manager
    def _connection_keys(self):
        return (
            "account",
            "user",
            "database",
            "warehouse",
            "role",
            "schema",
            "authenticator",
            "oauth_client_id",
            "query_tag",
            "client_session_keep_alive",
            "host",
            "port",
            "proxy_host",
            "proxy_port",
            "protocol",
            "connect_retries",
            "connect_timeout",
            "retry_on_database_errors",
            "retry_all",
            "insecure_mode",
            "reuse_connections",
            "s3_stage_vpce_dns_name",
            "platform_detection_timeout_seconds",
        )

    def auth_args(self):
        # Pull all of the optional authentication args for the connector,
        # let connector handle the actual arg validation
        result = {}
        if self.user:
            result["user"] = self.user
        if self.password:
            result["password"] = self.password
        if self.host:
            result["host"] = self.host
        if self.port:
            result["port"] = self.port
        if self.proxy_host:
            result["proxy_host"] = self.proxy_host
        if self.proxy_port:
            result["proxy_port"] = self.proxy_port
        if self.protocol:
            result["protocol"] = self.protocol
        if self.authenticator:
            result["authenticator"] = self.authenticator
            if self.authenticator == "oauth":
                token = self.token
                # if we have a client ID/client secret, the token is a refresh
                # token, not an access token
                if self.oauth_client_id and self.oauth_client_secret:
                    token = self._get_access_token()
                elif self.oauth_client_id:
                    warn_or_error(
                        AdapterEventWarning(
                            base_msg="Invalid profile: got an oauth_client_id, but not an oauth_client_secret!"
                        )
                    )
                elif self.oauth_client_secret:
                    warn_or_error(
                        AdapterEventWarning(
                            base_msg="Invalid profile: got an oauth_client_secret, but not an oauth_client_id!"
                        )
                    )

                result["token"] = token

            elif self.authenticator == "jwt":
                # If authenticator is 'jwt', then the 'token' value should be used
                # unmodified. We expose this as 'jwt' in the profile, but the value
                # passed into the snowflake.connect method should still be 'oauth'
                result["token"] = self.token
                result["authenticator"] = "oauth"

            # enable id token cache for linux
            result["client_store_temporary_credential"] = True
            # enable mfa token cache for linux
            result["client_request_mfa_token"] = True
        result["reuse_connections"] = self.reuse_connections
        result["private_key"] = self._get_private_key()
        return result

    def _get_access_token(self) -> str:
        if self.authenticator != "oauth":
            raise DbtInternalError("Can only get access tokens for oauth")
        missing = any(
            x is None for x in (self.oauth_client_id, self.oauth_client_secret, self.token)
        )
        if missing:
            raise DbtInternalError(
                "need a client ID a client secret, and a refresh token to get " "an access token"
            )

        # should the full url be a config item?
        token_url = _TOKEN_REQUEST_URL.format(self.account)
        # I think this is only used to redirect on success, which we ignore
        # (it does not have to match the integration's settings in snowflake)
        redirect_uri = "http://localhost:9999"
        data = {
            "grant_type": "refresh_token",
            "refresh_token": self.token,
            "redirect_uri": redirect_uri,
        }

        auth = base64.b64encode(
            f"{self.oauth_client_id}:{self.oauth_client_secret}".encode("ascii")
        ).decode("ascii")
        headers = {
            "Authorization": f"Basic {auth}",
            "Content-type": "application/x-www-form-urlencoded;charset=utf-8",
        }

        result_json = None
        max_iter = 20
        # Attempt to obtain JSON for 1 second before throwing an error
        for i in range(max_iter):
            result = requests.post(token_url, headers=headers, data=data)
            try:
                result_json = result.json()
                break
            except ValueError as e:
                message = result.text
                logger.debug(
                    f"Got a non-json response ({result.status_code}): \
                              {e}, message: {message}"
                )
                sleep(0.05)

        if result_json is None:
            raise DbtDatabaseError(
                f"""Did not receive valid json with access_token.
                                        Showing json response: {result_json}"""
            )
        elif "access_token" not in result_json:
            raise FailedToConnectError(
                "This error occurs when authentication has expired. "
                "Please reauth with your auth provider."
            )
        return result_json["access_token"]

    def _get_private_key(self) -> Optional[bytes]:
        """Get Snowflake private key by path, from a Base64 encoded DER bytestring or None."""
        if self.private_key and self.private_key_path:
            raise DbtConfigError("Cannot specify both `private_key`  and `private_key_path`")
        elif self.private_key:
            private_key = private_key_from_string(self.private_key, self.private_key_passphrase)
        elif self.private_key_path:
            private_key = private_key_from_file(self.private_key_path, self.private_key_passphrase)
        else:
            return None
        return snowflake_private_key(private_key)


class SnowflakeConnectionManager(SQLConnectionManager):
    TYPE = "snowflake"

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield
        except snowflake.connector.errors.ProgrammingError as e:
            msg = str(e)

            # A class of Snowflake errors -- such as a failure from attempting to merge
            # duplicate rows -- includes row values in the error message, i.e.
            # [12345, "col_a_value", "col_b_value", etc...]. We don't want to log potentially
            # sensitive user data.
            for regex_pattern, replacement_message in ERROR_REDACTION_PATTERNS.items():
                msg = re.sub(regex_pattern, replacement_message, msg)

            logger.debug("Snowflake query id: {}".format(e.sfqid))
            logger.debug("Snowflake error: {}".format(msg))

            if "Empty SQL statement" in msg:
                logger.debug("got empty sql statement, moving on")
            elif "This session does not have a current database" in msg:
                raise FailedToConnectError(
                    (
                        "{}\n\nThis error sometimes occurs when invalid "
                        "credentials are provided, or when your default role "
                        "does not have access to use the specified database. "
                        "Please double check your profile and try again."
                    ).format(msg)
                )
            else:
                raise DbtDatabaseError(msg)
        except Exception as e:
            if isinstance(e, snowflake.connector.errors.Error):
                logger.debug("Snowflake query id: {}".format(e.sfqid))

            logger.debug("Error running SQL: {}", sql)
            logger.debug("Rolling back transaction.")
            self.rollback_if_open()
            if isinstance(e, DbtRuntimeError):
                # during a sql query, an internal to dbt exception was raised.
                # this sounds a lot like a signal handler and probably has
                # useful information, so raise it without modification.
                raise
            raise DbtRuntimeError(str(e)) from e

    @classmethod
    def open(cls, connection):
        if connection.state == "open":
            logger.debug("Connection is already open, skipping open.")
            return connection

        creds = connection.credentials
        timeout = creds.connect_timeout

        def connect():
            session_parameters = {}

            if creds.query_tag:
                session_parameters.update({"QUERY_TAG": creds.query_tag})

            if creds.s3_stage_vpce_dns_name:
                session_parameters.update({"S3_STAGE_VPCE_DNS_NAME": creds.s3_stage_vpce_dns_name})
            handle = None

            # In replay mode, we won't connect to a real database at all, while
            # in record and diff modes we do, but insert an intermediate handle
            # object which monitors native connection activity.
            rec_mode = get_record_mode_from_env()
            handle = None
            if rec_mode != RecorderMode.REPLAY:
                handle = snowflake.connector.connect(
                    account=creds.account,
                    database=creds.database,
                    schema=creds.schema,
                    warehouse=creds.warehouse,
                    role=creds.role,
                    autocommit=True,
                    client_session_keep_alive=creds.client_session_keep_alive,
                    application="dbt",
                    insecure_mode=creds.insecure_mode,
                    platform_detection_timeout_seconds=creds.platform_detection_timeout_seconds,
                    session_parameters=session_parameters,
                    ocsp_root_certs_dict_lock_timeout=10,  # cert lock can cause deadlock without timeout, see https://github.com/snowflakedb/snowflake-connector-python/issues/2213
                    **creds.auth_args(),
                )

            if rec_mode is not None:
                # If using the record/replay mechanism, regardless of mode, we
                # use a wrapper.
                handle = SnowflakeRecordReplayHandle(handle, connection)

            return handle

        def exponential_backoff(attempt: int):
            return attempt * attempt

        retryable_exceptions = [
            InternalError,
            InternalServerError,
            ServiceUnavailableError,
            GatewayTimeoutError,
            RequestTimeoutError,
            BadGatewayError,
            OtherHTTPRetryableError,
            BindUploadError,
        ]
        # these two options are for backwards compatibility
        if creds.retry_all:
            retryable_exceptions = [Error]
        elif creds.retry_on_database_errors:
            retryable_exceptions.insert(0, DatabaseError)

        return cls.retry_connection(
            connection,
            connect=connect,
            logger=logger,
            retry_limit=creds.connect_retries,
            retry_timeout=timeout if timeout is not None else exponential_backoff,
            retryable_exceptions=retryable_exceptions,
        )

    def cancel(self, connection):
        handle = connection.handle
        sid = handle.session_id

        connection_name = connection.name

        sql = "select system$cancel_all_queries({})".format(sid)

        logger.debug("Cancelling query '{}' ({})".format(connection_name, sid))

        _, cursor = self.add_query(sql)
        res = cursor.fetchone()

        logger.debug("Cancel query '{}': {}".format(connection_name, res))

    @classmethod
    def get_response(cls, cursor) -> SnowflakeAdapterResponse:
        code = cursor.sqlstate

        if code is None:
            code = "SUCCESS"
        query_id = str(cursor.sfqid) if cursor.sfqid is not None else None

        # Extract DML stats from cursor.stats if available (snowflake-connector-python >= 4.2.0)
        rows_inserted = None
        rows_deleted = None
        rows_updated = None
        rows_duplicates = None
        if hasattr(cursor, "stats") and cursor.stats is not None:
            stats = cursor.stats
            rows_inserted = getattr(stats, "num_rows_inserted", None)
            rows_deleted = getattr(stats, "num_rows_deleted", None)
            rows_updated = getattr(stats, "num_rows_updated", None)
            rows_duplicates = getattr(stats, "num_dml_duplicates", None)

        # For CTAS and similar operations, rowcount is typically 1 (the success message row)
        # even when many rows are inserted. Use rows_inserted from stats for accurate reporting.
        rows_affected = cursor.rowcount
        if rows_inserted is not None and rows_inserted > 0:
            rows_affected = rows_inserted

        return SnowflakeAdapterResponse(
            _message="{} {}".format(code, rows_affected),
            rows_affected=rows_affected,
            code=code,
            query_id=query_id,
            rows_inserted=rows_inserted,
            rows_deleted=rows_deleted,
            rows_updated=rows_updated,
            rows_duplicates=rows_duplicates,
        )

    # disable transactional logic by default on Snowflake
    # except for DML statements where explicitly defined
    def add_begin_query(self, *args, **kwargs):
        pass

    def add_commit_query(self, *args, **kwargs):
        pass

    def begin(self):
        pass

    def commit(self):
        pass

    def clear_transaction(self):
        pass

    @classmethod
    def _split_queries(cls, sql):
        "Splits sql statements at semicolons into discrete queries"

        sql_s = str(sql)
        sql_buf = StringIO(sql_s)
        split_query = snowflake.connector.util_text.split_statements(sql_buf)
        return [part[0] for part in split_query]

    @staticmethod
    def _fix_rows(rows: Iterable[Iterable]) -> Iterable[Iterable]:
        # See note in process_results().
        for row in rows:
            fixed_row = []
            for col in row:
                if isinstance(col, datetime.datetime) and col.tzinfo:
                    offset = col.utcoffset()
                    assert offset is not None
                    offset_seconds = offset.total_seconds()
                    new_timezone = pytz.FixedOffset(int(offset_seconds // 60))
                    col = col.astimezone(tz=new_timezone)
                fixed_row.append(col)

            yield fixed_row

    @classmethod
    def process_results(cls, column_names, rows):
        # Override for Snowflake. The datetime objects returned by
        # snowflake-connector-python are not pickleable, so we need
        # to replace them with sane timezones.
        return super().process_results(column_names, cls._fix_rows(rows))

    def execute(
        self, sql: str, auto_begin: bool = False, fetch: bool = False, limit: Optional[int] = None
    ) -> Tuple[AdapterResponse, "agate.Table"]:
        # don't apply the query comment here
        # it will be applied after ';' queries are split
        from dbt_common.clients.agate_helper import empty_table

        _, cursor = self.add_query(sql, auto_begin)
        response = self.get_response(cursor)
        if fetch:
            table = self.get_result_from_cursor(cursor, limit)
        else:
            table = empty_table()
        return response, table

    def add_standard_query(self, sql: str, **kwargs) -> Tuple[Connection, Any]:
        # This is the happy path for a single query. Snowflake has a few odd behaviors that
        # require preprocessing within the 'add_query' method below.
        return super().add_query(self._add_query_comment(sql), **kwargs)

    def add_query(
        self,
        sql: str,
        auto_begin: bool = True,
        bindings: Optional[Any] = None,
        abridge_sql_log: bool = False,
        *args,
        **kwargs,
    ) -> Tuple[Connection, Any]:
        if bindings:
            # The snowflake connector is stricter than, e.g., psycopg2 -
            # which allows any iterable thing to be passed as a binding.
            bindings = tuple(bindings)

        stripped_queries = self._stripped_queries(sql)

        if set(query.lower() for query in stripped_queries).issubset({"begin;", "commit;"}):
            connection, cursor = self._add_begin_commit_only_queries(
                stripped_queries,
                auto_begin=auto_begin,
                bindings=bindings,
                abridge_sql_log=abridge_sql_log,
            )
        else:
            connection, cursor = self._add_standard_queries(
                stripped_queries,
                auto_begin=auto_begin,
                bindings=bindings,
                abridge_sql_log=abridge_sql_log,
            )

        if cursor is None:
            self._raise_cursor_not_found_error(sql)

        return connection, cursor

    def set_query_header(self, query_header_context: Dict[str, Any]) -> None:
        self.query_header = SnowflakeMacroQueryStringSetter(self.profile, query_header_context)

    def _stripped_queries(self, sql: str) -> List[str]:
        def strip_query(query):
            """
            hack -- after the last ';', remove comments and don't run
            empty queries. this avoids using exceptions as flow control,
            and also allows us to return the status of the last cursor
            """
            without_comments_re = re.compile(
                r"(\".*?\"|\'.*?\')|(/\*.*?\*/|--[^\r\n]*$)", re.MULTILINE
            )
            return re.sub(without_comments_re, "", query).strip()

        return [query for query in self._split_queries(sql) if strip_query(query) != ""]

    def _add_begin_commit_only_queries(
        self, queries: List[str], **kwargs
    ) -> Tuple[Connection, Any]:
        # if all we get is `begin;` and/or `commit;`
        # raise a warning, then run as standard queries to avoid an error downstream
        message = (
            "Explicit transactional logic should be used only to wrap "
            "DML logic (MERGE, DELETE, UPDATE, etc). The keywords BEGIN; and COMMIT; should "
            "be placed directly before and after your DML statement, rather than in separate "
            "statement calls or run_query() macros."
        )
        logger.warning(line_wrap_message(warning_tag(message)))

        for query in queries:
            connection, cursor = self.add_standard_query(query, **kwargs)
        return connection, cursor

    def _add_standard_queries(self, queries: List[str], **kwargs) -> Tuple[Connection, Any]:
        for query in queries:
            # Even though we turn off transactions by default for Snowflake,
            # the user/macro has passed them *explicitly*, probably to wrap a DML statement
            # This also has the effect of ignoring "commit" in the RunResult for this model
            # https://github.com/dbt-labs/dbt-snowflake/issues/147
            if query.lower() == "begin;":
                super().add_begin_query()
            elif query.lower() == "commit;":
                super().add_commit_query()
            else:
                # This adds a query comment to *every* statement
                # https://github.com/dbt-labs/dbt-snowflake/issues/140
                connection, cursor = self.add_standard_query(query, **kwargs)
        return connection, cursor

    def _raise_cursor_not_found_error(self, sql: str):
        conn = self.get_thread_connection()
        try:
            conn_name = conn.name
        except AttributeError:
            conn_name = None

        raise DbtRuntimeError(
            f"""Tried to run an empty query on model '{conn_name or "<None>"}'. If you are """
            f"""conditionally running\nsql, e.g. in a model hook, make """
            f"""sure your `else` clause contains valid sql!\n\n"""
            f"""Provided SQL:\n{sql}"""
        )

    def release(self):
        """Reuse connections by deferring release until adapter context manager in core
        resets adapters. This cleanup_all happens before Python teardown. Idle connections
        incur no costs while waiting in the connection pool."""
        if self.profile.credentials.reuse_connections:
            return
        super().release()

    @classmethod
    def data_type_code_to_name(cls, type_code: Union[int, str]) -> str:
        assert isinstance(type_code, int)
        return snowflake.connector.constants.FIELD_ID_TO_NAME[type_code]



from dbt.tests.util import run_dbt, run_dbt_and_capture
from dbt.tests.adapter.basic.test_incremental import (
    BaseIncremental,
    BaseIncrementalNotSchemaChange,
)


class TestBaseIncrementalNotSchemaChange(BaseIncrementalNotSchemaChange):
    pass


class TestIncrementalRunResultSnowflake(BaseIncremental):
    """Test to verify that incremental models return adapter response stats including
    DML statistics (rows_inserted, rows_updated, rows_deleted, rows_duplicates).

    This test checks both:
    1. Initial CTAS operation when the incremental table doesn't exist yet
    2. Incremental merge operation when the table already exists
    """

    def test_incremental(self, project):
        # seed command
        results = run_dbt(["seed"])
        assert len(results) == 2

        # First run with initial seed - this is a CTAS operation since the incremental table doesn't exist
        results, output = run_dbt_and_capture(["run", "--vars", "seed_name: base"])
        assert len(results) == 1

        # Verify CTAS adapter response includes DML stats
        adapter_response = results[0].adapter_response
        assert "code" in adapter_response
        assert adapter_response["code"] == "SUCCESS"

        # For CTAS, rows_inserted should equal the number of rows created (10 from base seed)
        assert "rows_inserted" in adapter_response
        assert (
            adapter_response["rows_inserted"] == 10
        ), f"Expected 10 rows inserted for CTAS, found {adapter_response['rows_inserted']}"

        # Other DML stats should be 0 for CTAS
        assert "rows_deleted" in adapter_response
        assert adapter_response["rows_deleted"] == 0
        assert "rows_updated" in adapter_response
        assert adapter_response["rows_updated"] == 0
        assert "rows_duplicates" in adapter_response
        assert adapter_response["rows_duplicates"] == 0

        # Verify stdout shows the row count for CTAS
        assert (
            "SUCCESS 10" in output
        ), f"Expected 'SUCCESS 10' in stdout for CTAS, but got: {output}"

        # Second run with additions - this is an incremental merge operation
        results, output = run_dbt_and_capture(["run", "--vars", "seed_name: added"])
        assert len(results) == 1

        # Verify incremental merge adapter response
        adapter_response = results[0].adapter_response
        rows_affected = adapter_response["rows_affected"]
        assert rows_affected == 10, f"Expected 10 rows changed, found {rows_affected}"

        # For incremental merge, rows_inserted should reflect new rows added
        assert "rows_inserted" in adapter_response
        assert (
            adapter_response["rows_inserted"] == 10
        ), f"Expected 10 rows inserted for merge, found {adapter_response['rows_inserted']}"

        # Verify stdout shows the row count for incremental merge
        assert (
            "SUCCESS 10" in output
        ), f"Expected 'SUCCESS 10' in stdout for incremental merge, but got: {output}"



from dbt.tests.adapter.dbt_show.test_dbt_show import (
    BaseShowSqlHeader,
    BaseShowLimit,
    BaseShowDoesNotHandleDoubleLimit,
)


class TestSnowflakeShowLimit(BaseShowLimit):
    pass


class TestSnowflakeShowSqlHeader(BaseShowSqlHeader):
    pass


class TestSnowflakeShowDoesNotHandleDoubleLimit(BaseShowDoesNotHandleDoubleLimit):
    DATABASE_ERROR_MESSAGE = "unexpected 'limit'"



import pytest

from dbt.tests.util import (
    run_dbt,
)

_MODELS__incremental_model = """
{{ config(
    materialized='incremental',
    unique_key='id'
) }}

with data as (
    SELECT $1 id, $2 name FROM (
        VALUES (1, 'one'), (2, 'two'), (3, 'three'), (1, 'one')
    )
)
select * from data
"""


class TestRowValuesNotInExceptions:
    @pytest.fixture(scope="class")
    def models(self):
        return {"model.sql": _MODELS__incremental_model}

    def test_row_values_were_scrubbed_from_duplicate_merge_exception(self, project):
        result = run_dbt(["run", "-s", "model"])
        assert len(result) == 1

        result = run_dbt(["run", "-s", "model"], expect_pass=False)
        assert len(result) == 1
        assert "Row Values: [redacted]" in result[0].message
        assert "'one'" not in result[0].message



from dbt.tests.util import run_dbt_and_capture
import pytest


ANONYMOUS_USAGE_MESSAGE = """
sys._xoptions['snowflake_partner_attribution'].append("dbtLabs_dbtPython")
""".strip()


MY_PYTHON_MODEL = """
import pandas

def model(dbt, session):
    dbt.config(materialized='table')
    data = [[1,2]] * 10
    return pandas.DataFrame(data, columns=['test', 'test2'])
"""


class AnonymousUsageStatsBase:
    @pytest.fixture(scope="class")
    def models(self):
        return {"my_python_model.py": MY_PYTHON_MODEL}


class TestAnonymousUsageStatsOn(AnonymousUsageStatsBase):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"flags": {"send_anonymous_usage_stats": True}}

    def test_stats_get_sent(self, project):
        _, logs = run_dbt_and_capture(["--debug", "run"])
        assert ANONYMOUS_USAGE_MESSAGE in logs


class TestAnonymousUsageStatsOff(AnonymousUsageStatsBase):
    @pytest.fixture(scope="class")
    def project_config_update(self, dbt_profile_target):
        return {"flags": {"send_anonymous_usage_stats": False}}

    def test_stats_do_not_get_sent(self, project):
        _, logs = run_dbt_and_capture(["--debug", "run"])
        assert ANONYMOUS_USAGE_MESSAGE not in logs



"""
Results:

| method        | project_size | reuse_connections | unsafe_skip_rsa_key_validation | duration |
|---------------|--------------|-------------------|--------------------------------|----------|
| User Password |        1,000 | False             | -                              |  234.09s |
| User Password |        1,000 | True              | -                              |   78.34s |
| Key Pair      |        1,000 | False             | False                          |  271.47s |
| Key Pair      |        1,000 | False             | True                           |  275.73s |
| Key Pair      |        1,000 | True              | False                          |   63.69s |
| Key Pair      |        1,000 | True              | True                           |   73.45s |

Notes:
- run locally on MacOS, single threaded
- `unsafe_skip_rsa_key_validation` only applies to the Key Pair auth method
- `unsafe_skip_rsa_key_validation=True` was tested by updating the relevant `cryptography` calls directly as it is not a user configuration
- since the models are all views, time differences should be viewed as absolute differences, e.g.:
    - this: (271.47s - 63.69s) / 1,000 models = 208ms improvement
    - NOT this: 1 - (63.69s / 271.47s) = 76.7% improvement
"""

from datetime import datetime
import os

from dbt.tests.util import run_dbt
import pytest


SEED = """
id,value
1,a
2,b
3,c
""".strip()


MODEL = """
select * from {{ ref("my_seed") }}
"""


class Scenario:
    """
    Runs a full load test. The test can be configured to run an arbitrary number of models.

    To use this test, configure the test by setting `project_size` and/or `expected_duration`.
    """

    auth_method: str
    project_size: int = 1
    reuse_connections: bool = False

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"my_seed.csv": SEED}

    @pytest.fixture(scope="class")
    def models(self):
        return {f"my_model_{i}.sql": MODEL for i in range(self.project_size)}

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        run_dbt(["seed"])

        start = datetime.now()
        yield
        end = datetime.now()

        duration = (end - start).total_seconds()
        print(f"Run took: {duration} seconds")

    @pytest.fixture(scope="class")
    def dbt_profile_target(self, auth_params):
        yield {
            "type": "snowflake",
            "threads": 4,
            "account": os.getenv("SNOWFLAKE_TEST_ACCOUNT"),
            "database": os.getenv("SNOWFLAKE_TEST_DATABASE"),
            "warehouse": os.getenv("SNOWFLAKE_TEST_WAREHOUSE"),
            "user": os.getenv("SNOWFLAKE_TEST_USER"),
            "reuse_connections": self.reuse_connections,
            **auth_params,
        }

    @pytest.fixture(scope="class")
    def auth_params(self):

        if self.auth_method == "user_password":
            yield {"password": os.getenv("SNOWFLAKE_TEST_PASSWORD")}

        elif self.auth_method == "key_pair":
            """
            This connection method uses key pair auth.
            Follow the instructions here to setup key pair authentication for your test user:
            https://docs.snowflake.com/en/user-guide/key-pair-auth
            """
            yield {
                "private_key": os.getenv("SNOWFLAKE_TEST_PRIVATE_KEY"),
                "private_key_passphrase": os.getenv("SNOWFLAKE_TEST_PRIVATE_KEY_PASSPHRASE"),
            }

        else:
            raise ValueError(
                f"`auth_method` must be one of `user_password` or `key_pair`, received: {self.auth_method}"
            )

    def test_scenario(self, project):
        run_dbt(["run"])


class TestUserPasswordAuth(Scenario):
    auth_method = "user_password"
    project_size = 1_000
    reuse_connections = False


class TestUserPasswordAuthReuseConnections(Scenario):
    auth_method = "user_password"
    project_size = 1_000
    reuse_connections = True


class TestKeyPairAuth(Scenario):
    auth_method = "key_pair"
    project_size = 1_000
    reuse_connections = False


class TestKeyPairAuthReuseConnections(Scenario):
    auth_method = "key_pair"
    project_size = 1_000
    reuse_connections = True



import pytest

from pathlib import Path

from dbt.tests.util import run_dbt, rm_file, write_file, check_relations_equal

from dbt.tests.adapter.simple_copy.test_simple_copy import (
    SimpleCopySetup,
    SimpleCopyBase,
    EmptyModelsArentRunBase,
)

from tests.functional.adapter.simple_copy.fixtures import (
    _MODELS__INCREMENTAL_OVERWRITE,
    _MODELS__INCREMENTAL_UPDATE_COLS,
    _SEEDS__SEED_MERGE_EXPECTED,
    _SEEDS__SEED_MERGE_INITIAL,
    _SEEDS__SEED_MERGE_UPDATE,
    _SEEDS__SEED_UPDATE,
    _TESTS__GET_RELATION_QUOTING,
)


class TestSimpleCopyBase(SimpleCopyBase):
    @pytest.fixture(scope="class")
    def tests(self):
        return {"get_relation_test.sql": _TESTS__GET_RELATION_QUOTING}

    def test_simple_copy(self, project):
        super().test_simple_copy(project)
        run_dbt(["test"])


class TestSimpleCopyBaseQuotingOff(SimpleCopyBase):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"quoting": {"identifier": False}}

    @pytest.fixture(scope="class")
    def tests(self):
        return {"get_relation_test.sql": _TESTS__GET_RELATION_QUOTING}

    def test_simple_copy_quoting_off(self, project):
        super().test_simple_copy(project)
        run_dbt(["test"])

    @pytest.mark.skip(reason="Already run and test case above; no need to run again")
    def test_simple_copy(self):
        pass

    @pytest.mark.skip(reason="Already run and test case above; no need to run again")
    def test_simple_copy_with_materialized_views(self):
        pass


class TestEmptyModelsArentRun(EmptyModelsArentRunBase):
    pass


file_append = """
quoting:
    identifier: true
"""


class TestSimpleCopyBaseQuotingSwitch(SimpleCopySetup):
    @pytest.fixture(scope="class")
    def tests(self):
        return {"get_relation_test.sql": _TESTS__GET_RELATION_QUOTING}

    def test_seed_quoting_switch(self, project):
        results = run_dbt(["seed"])
        assert len(results) == 1

        # Update seed file
        main_seed_file = project.project_root / Path("seeds") / Path("seed.csv")
        rm_file(main_seed_file)
        write_file(_SEEDS__SEED_UPDATE, main_seed_file)

        # Change the profile temporarily
        dbt_project_yml = project.project_root / Path("dbt_project.yml")
        with open(dbt_project_yml, "r+") as f:
            dbt_project_yml_contents = f.read()
            f.write(file_append)
        run_dbt(["seed"], expect_pass=False)

        with open(dbt_project_yml, "w") as f:
            f.write(dbt_project_yml_contents)
        run_dbt(["test"])


inc_strat_yml = """
models:
    incremental_strategy: "delete+insert"
"""


class TestSnowflakeIncrementalOverwrite(SimpleCopySetup):
    @pytest.fixture(scope="class")
    def models(self):
        return {"incremental_overwrite.sql": _MODELS__INCREMENTAL_OVERWRITE}

    def test__snowflake__incremental_overwrite(self, project):
        results = run_dbt(["seed"])
        assert len(results) == 1
        results = run_dbt(["run"])
        assert len(results) == 1

        # Fails using 'merge' strategy because there's a duplicate 'id'
        results = run_dbt(["run"], expect_pass=False)
        assert len(results) == 1

        # Setting the incremental_strategy should make this succeed
        dbt_project_yml = project.project_root / Path("dbt_project.yml")
        with open(dbt_project_yml, "a") as f:
            f.write(inc_strat_yml)

        results = run_dbt(["run"])
        assert len(results) == 1


class TestIncrementalMergeColumns(SimpleCopySetup):
    @pytest.fixture(scope="class")
    def models(self):
        return {"incremental_update_cols.sql": _MODELS__INCREMENTAL_UPDATE_COLS}

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": _SEEDS__SEED_MERGE_INITIAL,
            "expected_result.csv": _SEEDS__SEED_MERGE_EXPECTED,
        }

    def seed_and_run(self):
        run_dbt(["seed"])
        run_dbt(["run"])

    def test__snowflake__incremental_merge_columns(self, project):
        self.seed_and_run()

        main_seed_file = project.project_root / Path("seeds") / Path("seed.csv")
        rm_file(main_seed_file)
        write_file(_SEEDS__SEED_MERGE_UPDATE, main_seed_file)
        self.seed_and_run()
        check_relations_equal(project.adapter, ["incremental_update_cols", "expected_result"])



import pytest
from dbt.tests.util import check_table_does_exist, run_dbt

_SEED_CSV = """
id,first_name,last_name,email,product_id
1,Jack,Hunter,jhunter0@foo.bar,1
2,Kathryn,Walker,kwalker1@foo.bar,1
3,Gerald,Ryan,gryan2@foo.bar,3
4,Jack,Hunter,jhunter1@foo.bar,4
5,Kathryn,Walker,kwalker2@foo.bar,5
6,Gerald,Ryan,gryan3@foo.bar,6
""".lstrip()


class TestClusterBy:
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"seed.csv": _SEED_CSV}

    @pytest.fixture(scope="class")
    def models(self, dbt_profile_target):
        warehouse_name = dbt_profile_target["warehouse"]

        _DYNAMIC_TABLE_1_SQL = f"""
        {{{{ config(materialized='dynamic_table', snowflake_warehouse='{warehouse_name}', target_lag='1 minute') }}}}
        select * from {{{{ ref('seed') }}}}
        """.lstrip()

        _DYNAMIC_TABLE_2_SQL = f"""
        {{{{ config(materialized='dynamic_table', cluster_by=['last_name'], snowflake_warehouse='{warehouse_name}', target_lag='1 minute') }}}}
        select * from {{{{ ref('dynamic_table_1') }}}}
        """.lstrip()

        _DYNAMIC_TABLE_3_SQL = f"""
        {{{{ config(materialized='dynamic_table', cluster_by=['last_name', 'first_name'], snowflake_warehouse='{warehouse_name}', target_lag='1 minute') }}}}
        select
            last_name,
            first_name,
            count(*) as count
        from {{{{ ref('seed') }}}}
        group by 1, 2
        """.lstrip()

        _DYNAMIC_TABLE_4_SQL = f"""
        {{{{ config(materialized='dynamic_table', cluster_by=['last_name', 'product_id % 3'], snowflake_warehouse='{warehouse_name}', target_lag='1 minute') }}}}
        select
            last_name,
            first_name,
            product_id,
            count(*) as count
        from {{{{ ref('seed') }}}}
        group by 1, 2, 3
        """.lstrip()

        return {
            "dynamic_table_1.sql": _DYNAMIC_TABLE_1_SQL,
            "dynamic_table_2.sql": _DYNAMIC_TABLE_2_SQL,
            "dynamic_table_3.sql": _DYNAMIC_TABLE_3_SQL,
            "dynamic_table_4.sql": _DYNAMIC_TABLE_4_SQL,
        }

    def test_snowflake_dynamic_table_cluster_by(self, project):

        run_dbt(["seed"])

        db_with_schema = f"{project.database}.{project.test_schema}"

        check_table_does_exist(
            project.adapter, f"{db_with_schema}.{self._available_models_in_setup()['seed_table']}"
        )

        run_dbt()

        # Check that all dynamic tables exist
        check_table_does_exist(
            project.adapter,
            f"{db_with_schema}.{self._available_models_in_setup()['dynamic_table_1']}",
        )
        check_table_does_exist(
            project.adapter,
            f"{db_with_schema}.{self._available_models_in_setup()['dynamic_table_2']}",
        )
        check_table_does_exist(
            project.adapter,
            f"{db_with_schema}.{self._available_models_in_setup()['dynamic_table_3']}",
        )
        check_table_does_exist(
            project.adapter,
            f"{db_with_schema}.{self._available_models_in_setup()['dynamic_table_4']}",
        )

        with project.adapter.connection_named("__test"):
            # Check if cluster_by is applied to dynamic_table_2 (should cluster by last_name)
            cluster_by = self._get_dynamic_table_ddl(
                project, self._available_models_in_setup()["dynamic_table_2"]
            )
            assert "CLUSTER BY (LAST_NAME)" in cluster_by.upper()

            # Check if cluster_by is applied to dynamic_table_3 (should cluster by last_name, first_name)
            cluster_by = self._get_dynamic_table_ddl(
                project, self._available_models_in_setup()["dynamic_table_3"]
            )
            assert "CLUSTER BY (LAST_NAME, FIRST_NAME)" in cluster_by.upper()

            # Check if cluster_by is applied to dynamic_table_4 (should cluster by last_name, product_id % 3)
            cluster_by = self._get_dynamic_table_ddl(
                project, self._available_models_in_setup()["dynamic_table_4"]
            )
            assert "CLUSTER BY (LAST_NAME, PRODUCT_ID % 3)" in cluster_by.upper()

    def _get_dynamic_table_ddl(self, project, table_name: str) -> str:
        ddl_query = f"SELECT GET_DDL('DYNAMIC_TABLE', '{project.database}.{project.test_schema}.{table_name}')"
        ddl = project.run_sql(ddl_query, fetch="one")
        return ddl[0]

    def _available_models_in_setup(self) -> dict[str, str]:
        return dict(
            seed_table="SEED",
            dynamic_table_1="DYNAMIC_TABLE_1",
            dynamic_table_2="DYNAMIC_TABLE_2",
            dynamic_table_3="DYNAMIC_TABLE_3",
            dynamic_table_4="DYNAMIC_TABLE_4",
        )



from dbt.tests.adapter.caching.test_caching import (
    BaseCachingLowercaseModel,
    BaseCachingUppercaseModel,
    BaseCachingSelectedSchemaOnly,
)


class TestCachingLowerCaseModel(BaseCachingLowercaseModel):
    pass


class TestCachingUppercaseModel(BaseCachingUppercaseModel):
    pass


class TestCachingSelectedSchemaOnly(BaseCachingSelectedSchemaOnly):
    pass



import os
import re
import pytest

from dbt.tests.adapter.catalog_integrations.test_catalog_integration import (
    BaseCatalogIntegrationValidation,
)
from dbt.tests.util import run_dbt, run_dbt_and_capture, write_file


def get_cleaned_model_ddl_from_file(file_name: str) -> str:
    with open(f"target/run/test/models/{file_name}", "r") as ddl_file:
        return re.sub(r"\s+", " ", ddl_file.read())


_SEED_PARTITION_BY = """
order_id,order_date,region,amount
1001,2025-10-01,North,120.50
1002,2025-10-01,North,85.00
1003,2025-10-01,West,42.10
1004,2025-10-02,West,110.00
1005,2025-10-02,East,67.30
1006,2025-10-02,East,134.40
1007,2025-10-03,South,210.00
1008,2025-10-03,South,95.00
1009,2025-10-03,North,55.00
1010,2025-10-04,North,33.30
1011,2025-10-04,West,145.20
1012,2025-10-04,East,60.00
1013,2025-10-05,West,88.90
1014,2025-10-05,South,75.50
1015,2025-10-05,East,99.00
1016,2025-10-06,North,122.00
1017,2025-10-06,South,140.00
1018,2025-10-06,West,57.50
1019,2025-10-07,North,200.00
1020,2025-10-07,East,180.00
""".strip()

_MODEL_BUILTIN_TABLE = """
{{
  config(
    materialized='table',
    partition_by=['order_date']
) }}

select
  order_date,
  region,
  count(*) as num_orders,
  sum(amount) as total_sales,
  avg(amount) as avg_order_value
from {{ ref('seed') }}
group by 1, 2
order by 1, 2
"""

_MODEL_ICEBERG_BUILTIN_TABLE = """
{{
  config(
    materialized='table',
    catalog_name='iceberg_snowflake_managed_catalog',
    partition_by=['order_date']
) }}

select
  order_date,
  region,
  count(*) as num_orders,
  sum(amount) as total_sales,
  avg(amount) as avg_order_value
from {{ ref('seed') }}
group by 1, 2
order by 1, 2
"""

_MODEL_ICEBERG_REST_TABLE = """
{{
  config(
    materialized='table',
    catalog_name='iceberg_rest_catalog',
    partition_by=['order_date']
) }}

select
  order_date,
  region,
  count(*) as num_orders,
  sum(amount) as total_sales,
  avg(amount) as avg_order_value
from {{ ref('seed') }}
group by 1, 2
order by 1, 2
"""

_MODEL_GLUE_ICEBERG_REST_TABLE = """
{{
  config(
    materialized='table',
    catalog_name='glue_iceberg_rest_catalog',
    partition_by=['order_date']
) }}

select
  order_date,
  region,
  count(*) as num_orders,
  sum(amount) as total_sales,
  avg(amount) as avg_order_value
from {{ ref('seed') }}
group by 1, 2
order by 1, 2
"""


class TestIcebergPartitionBy(BaseCatalogIntegrationValidation):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": _SEED_PARTITION_BY,
        }

    @pytest.fixture(scope="class")
    def catalogs(self):
        return {
            "catalogs": [
                {
                    "name": "iceberg_snowflake_managed_catalog",
                    "active_write_integration": "iceberg_managed_test",
                    "write_integrations": [
                        {
                            "name": "iceberg_managed_test",
                            "external_volume": "s3_iceberg_snow",
                            "catalog_type": "built_in",
                            "table_format": "iceberg",
                            "adapter_properties": {
                                # No catalog_linked_database_type means standard CTAS is used
                                "change_tracking": "true",
                            },
                        }
                    ],
                },
                {
                    "name": "iceberg_rest_catalog",
                    "active_write_integration": "iceberg_rest_test",
                    "write_integrations": [
                        {
                            "name": "iceberg_rest_test",
                            "catalog_type": "iceberg_rest",
                            "table_format": "iceberg",
                            "adapter_properties": {
                                "catalog_linked_database": os.getenv(
                                    "SNOWFLAKE_TEST_CATALOG_LINKED_DATABASE"
                                ),
                                "max_data_extension_time_in_days": 1,
                                "target_file_size": "AUTO",
                                "auto_refresh": "true",
                            },
                        }
                    ],
                },
                {
                    "name": "glue_iceberg_rest_catalog",
                    "active_write_integration": "glue_iceberg_rest_catalog_integration",
                    "write_integrations": [
                        {
                            "name": "glue_iceberg_rest_catalog_integration",
                            "catalog_type": "iceberg_rest",
                            "table_format": "iceberg",
                            "adapter_properties": {
                                "catalog_linked_database": os.getenv(
                                    "SNOWFLAKE_TEST_CATALOG_LINKED_DATABASE_GLUE"
                                ),
                                "catalog_linked_database_type": "glue",  # Glue requires 4-step process
                                "max_data_extension_time_in_days": 1,
                                "target_file_size": "AUTO",
                                "auto_refresh": "true",
                            },
                        }
                    ],
                },
            ]
        }

    # can only use alphanumeric characters in schema names in catalog linked databases
    # until it's GA
    @pytest.fixture(scope="class")
    def unique_schema(self, request, prefix) -> str:
        test_file = request.module.__name__
        # We only want the last part of the name
        test_file = test_file.split(".")[-1]
        unique_schema = f"{prefix}_{test_file}"
        return unique_schema.replace("_", "")


class TestPartitionByIgnoredIfNotIceberg(TestIcebergPartitionBy):
    @pytest.fixture(scope="function", autouse=True)
    def setup_class(self, project):
        run_dbt(["seed"])
        yield

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "standard_table.sql": _MODEL_BUILTIN_TABLE,
        }

    def test_partition_by_ignored_if_not_iceberg(self, project, setup_class):
        run_dbt(["run"])
        standard_table_sql = get_cleaned_model_ddl_from_file("standard_table.sql")
        assert "partition by" not in standard_table_sql


class TestPartitionByIcebergBuiltinCatalog(TestIcebergPartitionBy):
    @pytest.fixture(scope="function", autouse=True)
    def setup_class(self, project):
        run_dbt(["seed"])
        yield

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "builtin_table.sql": _MODEL_ICEBERG_BUILTIN_TABLE,
        }

    def test_partition_by_iceberg_builtin_catalog(self, project, setup_class):
        run_dbt(["run"])
        iceberg_sql = get_cleaned_model_ddl_from_file("builtin_table.sql")
        assert "partition by (order_date)" in iceberg_sql


class TestPartitionByIcebergRestCatalog(TestIcebergPartitionBy):
    @pytest.fixture(scope="function", autouse=True)
    def setup_class(self, project):
        run_dbt(["seed"])
        yield

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "rest_table.sql": _MODEL_ICEBERG_REST_TABLE,
        }

    def test_partition_by_iceberg_rest_catalog(self, project, setup_class):
        run_dbt(["run"])
        iceberg_sql = get_cleaned_model_ddl_from_file("rest_table.sql")
        assert "partition by (order_date)" in iceberg_sql


class TestPartitionByIcebergRestGlueCatalog(TestIcebergPartitionBy):
    # For some reason, when using Glue, `dbt seed` doesn't create the seed schema.
    @pytest.fixture(scope="class", autouse=True)
    def setup_glue_seed(self, project):
        """Pre-create schema with quoted lowercase identifier for Glue CLD"""
        adapter = project.adapter
        seed_database = os.getenv("SNOWFLAKE_TEST_DATABASE")
        schema_name = project.test_schema.lower()

        # Create schema with quoted identifier to preserve lowercase
        create_schema_sql = f'CREATE SCHEMA IF NOT EXISTS {seed_database}."{schema_name}"'
        adapter.execute(create_schema_sql, fetch=False)

        yield

        # Cleanup: drop schema after test
        drop_schema_sql = f'DROP SCHEMA IF EXISTS {seed_database}."{schema_name}"'
        try:
            adapter.execute(drop_schema_sql, fetch=False)
        except:
            pass  # Ignore cleanup errors

    @pytest.fixture(scope="function", autouse=True)
    def setup_class(self, project):
        run_dbt(["seed", "--log-level", "debug"])
        yield

    @pytest.fixture(scope="class")
    def project_config_update(self):
        # Force quoting for Glue CLD compatibility
        return {
            "quoting": {
                "database": False,
                "schema": True,
                "identifier": True,
            }
        }

    # AWS Glue requires lowercase identifiers and alphanumeric characters only
    @pytest.fixture(scope="class")
    def unique_schema(self, request, prefix) -> str:
        test_file = request.module.__name__
        # We only want the last part of the name
        test_file = test_file.split(".")[-1]
        unique_schema = f"{prefix}_{test_file}_glue"
        # Remove underscores and convert to lowercase for Glue compatibility
        return unique_schema.replace("_", "").lower()

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "glue_table.sql": _MODEL_GLUE_ICEBERG_REST_TABLE,
        }

    def test_partition_by_glue_iceberg_rest_catalog(self, project, setup_class):
        run_dbt(["run", "--log-level", "debug"])
        iceberg_sql = get_cleaned_model_ddl_from_file("glue_table.sql")
        assert 'partition by ("order_date")' in iceberg_sql



import dataclasses
from typing import Any, Optional

from dbt_common.record import Record, Recorder


@dataclasses.dataclass
class CursorGetStatsParams:
    connection_name: str


@dataclasses.dataclass
class CursorGetStatsResult:
    """Captures the stats from SnowflakeCursor.stats (QueryResultStats).

    These stats are available in snowflake-connector-python >= 4.2.0 and provide
    detailed DML operation information.
    """

    num_rows_inserted: Optional[int]
    num_rows_deleted: Optional[int]
    num_rows_updated: Optional[int]
    num_dml_duplicates: Optional[int]


class StatsProxy:
    """A proxy object that mimics the QueryResultStats interface for replay mode."""

    def __init__(self, result: Optional[CursorGetStatsResult]) -> None:
        self._result = result

    @property
    def num_rows_inserted(self) -> Optional[int]:
        return self._result.num_rows_inserted if self._result else None

    @property
    def num_rows_deleted(self) -> Optional[int]:
        return self._result.num_rows_deleted if self._result else None

    @property
    def num_rows_updated(self) -> Optional[int]:
        return self._result.num_rows_updated if self._result else None

    @property
    def num_dml_duplicates(self) -> Optional[int]:
        return self._result.num_dml_duplicates if self._result else None


@Recorder.register_record_type
class CursorGetStatsRecord(Record):
    params_cls = CursorGetStatsParams
    result_cls = CursorGetStatsResult
    group = "Database"

    @classmethod
    def _record_result(cls, result: Any) -> Optional[CursorGetStatsResult]:
        """Convert the native stats object to our result dataclass."""
        if result is None:
            return None
        return CursorGetStatsResult(
            num_rows_inserted=getattr(result, "num_rows_inserted", None),
            num_rows_deleted=getattr(result, "num_rows_deleted", None),
            num_rows_updated=getattr(result, "num_rows_updated", None),
            num_dml_duplicates=getattr(result, "num_dml_duplicates", None),
        )

    @classmethod
    def _replay_result(cls, result: Optional[CursorGetStatsResult]) -> Optional[StatsProxy]:
        """Convert our result dataclass back to a stats-like object for replay."""
        if result is None:
            return None
        return StatsProxy(result)



from dbt.tests.adapter.simple_snapshot.test_ephemeral_snapshot_hard_deletes import (
    BaseSnapshotEphemeralHardDeletes,
)


class TestSnapshotEphemeralHardDeletes(BaseSnapshotEphemeralHardDeletes):
    pass
